#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
``patato-draw-roi``: A script to draw regions of interest interactively on datasets..
"""

import argparse
import glob
from collections import namedtuple
from os.path import join, split

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Slider, Button, RangeSlider, PolygonSelector
from .. import ImageSequence, ROI, PAData
from ..utils import sort_key
from scipy.optimize import least_squares

if matplotlib.get_backend() == "MacOSX":
    matplotlib.use("TkAgg")

from ..utils.roi_operations import ROI_NAMES, REGION_COLOURS, close_loop

# TODO: Temporary fix for ?
import os

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"


def circ_loss(x, points):
    R = x[0]
    x_0 = x[1]
    y_0 = x[2]
    return np.sum(((points[:, 0] - x_0) ** 2 / R ** 2 + (points[:, 1] - y_0) ** 2 / R ** 2 - 1) ** 2)


def get_circle_points(r, x_0, y_0, n_points=50):
    thetas = np.linspace(0, 2 * np.pi, n_points, False)
    return np.array([np.sin(thetas) * r + x_0, np.cos(thetas) * r + y_0]).T


def fitcirc(points):
    points = np.array(points)
    x_0 = np.mean(points[:, 0])
    y_0 = np.mean(points[:, 1])
    r_0 = np.sqrt(np.mean((points[:, 0] - x_0) ** 2 + (points[:, 1] - y_0) ** 2))
    fit = least_squares(circ_loss, x0=np.array([r_0, x_0, y_0]), args=(points,))
    return get_circle_points(*fit.x)


positions = ["left", "right", "full", "", "radial", "ulnar", "superficial"]


def init_argparse():
    parser = argparse.ArgumentParser(description="Process MSOT Data.")
    parser.add_argument('input', type=str, help="Data Folder")
    parser.add_argument('-n', '--name', type=str, help="ROI Name",
                        choices=ROI_NAMES, default="unnamed")
    parser.add_argument('-p', '--position', type=str, help="ROI Location",
                        choices=positions, default="")
    parser.add_argument('-f', '--filter', type=str, help="Choose scan", default=None)
    parser.add_argument('-r', '--recon', type=int, help="Reconstruction number", default=0)
    parser.add_argument('-c', '--clear', type=bool, help="Clear Rois", default=False)
    parser.add_argument('-fn', '--filtername', type=str, help="Choose scan name filter", default=None)
    parser.add_argument('-thb', '--drawthb', type=bool, help="Draw on THb", default=False)
    parser.add_argument('-s', '--drawsigso2', type=bool, help="Draw on SigSo2", default=False)
    parser.add_argument('-u', '--drawultrasound', type=bool, help="Draw on Ultrasound", default=False)
    parser.add_argument('-o', '--overlaythb', type=bool, help="Draw on THb", default=False)
    parser.add_argument('-t', '--overlaythresh', type=float, help="THb threshold", default=80.)
    parser.add_argument('-fc', '--fitcircles', type=bool, help="Fit Circles To Rois", default=False)
    parser.add_argument('-l', '--logoverlay', type=bool, help="Log of overlay", default=False)
    parser.add_argument('-g', '--start_scan_number', type=int, help="Start drawing at scan n", default=None)
    parser.add_argument('-i', '--interpolation', type=bool, help="Interpolate between ROI slices", default=False)
    parser.add_argument('-w', '--wavelength', type=float, help="Default wavelength", default=0)
    parser.add_argument("--loadall", type=bool, default=False, help="Preload data to memory.")
    parser.add_argument("--drawallrois", type=bool, default=False, help="Display all rois on all slices.")
    return parser


class ROIDrawer:
    def update(self, _=None):
        if self.previous_frame_number != self.i_index:
            self.selection_n = -1
        if self.selection_n == -1:
            self.edit_btn.label.set_text("No ROI Selected")
            self.edit_btn.set_active(False)
            self.delete_btn.label.set_text("No ROI Selected")
            self.delete_btn.set_active(False)
        else:
            self.edit_btn.label.set_text("Change ROI Name")
            self.edit_btn.set_active(True)
            self.delete_btn.label.set_text("Delete ROI")
            self.delete_btn.set_active(True)
        self.previous_frame_number = self.i_index

        # Update the images that are shown
        self._draw_image()
        # self.main_image.set_data(np.squeeze(self.image_data[self.i_index, self.j_index].raw_data))
        # if self.overlay_data is not None:
        #     self.overlay_image.set_data(np.squeeze(self.overlay_data[self.i_index, self.j_index].raw_data))

        # remove the roi lines
        for r in self.roi_lines:
            line = r.pop(0)
            line.remove()

        self.roi_lines = []
        self.roi_names = []

        # Plot the regions.
        roi_number = 0
        for (r, n), roi_data in self.pa_data.get_rois(interpolate=self.interpolation).items():
            colour = "C0"
            for k, c in zip(ROI_NAMES, REGION_COLOURS):
                if k in r:
                    colour = c

            roi_points = roi_data.points

            if np.isclose(self.match_frames[self.i_index, self.j_index],
                          roi_data.attributes.get(self.frame_type, 1.0)) or self.draw_all_rois:
                roi_close = close_loop(roi_points)
                roi_plot = self.image_axis.plot(roi_close[:, 0], roi_close[:, 1], picker=True,
                                                label=r + "/" + n, c=colour, scalex=False,
                                                scaley=False)
                # Show the ROI on the appropriate frame.
                if roi_number == self.selection_n:
                    roi_plot[0].set_linestyle("dashed")
                    roi_plot[0].set_zorder(100)
                self.roi_lines.append(roi_plot)
                self.roi_names.append(r + "/" + n)
                roi_number += 1

        frame_number = 0
        # Draw the ROI names
        for lin, nam in zip(self.roi_lines, self.roi_names):
            if frame_number >= len(self.legend):
                self.legend.append(
                    self.fig.text(self.legend_x_0, self.legend_y_0 + self.legend_y,
                                  nam, color=lin[0].get_color(), picker=True))
                self.legend_y += self.legend_dy
            else:
                self.legend[frame_number].set_text(nam)
                self.legend[frame_number].set_color(lin[0].get_color())
            if frame_number == self.selection_n:
                self.legend[frame_number].set_fontweight("extra bold")
            else:
                self.legend[frame_number].set_fontweight("normal")
            frame_number += 1

        while frame_number < len(self.legend):
            self.legend[frame_number].set_text("")
            frame_number += 1

        # Add some labels.
        z = self.pa_data.get_z_positions()[self.i_index, self.j_index]
        self.z_text.set_text(f"z = {z:.2f} mm")
        wavelength = self.pa_data.get_wavelengths()[self.j_index]
        self.wl_text.set_text(f"wavelength = {wavelength:.0f} nm")

        # Reset the colour limits
        a, b = self.ax_clims.get_xlim()
        self.ax_clims.set_xlim((min([np.nanmin(self.main_image.get_array()), a]),
                                max([np.nanmax(self.main_image.get_array()), b])))

        # Redraw
        if plt.get_backend() == "MacOSX":
            self.fig.canvas.draw_idle()
        else:
            self.fig.canvas.draw()

    def on_pick(self, event):
        if not self.drawing:
            line = event.artist
            try:
                name = line.get_text()
            except AttributeError:
                name = line.get_label()
            if self.selection_n != self.roi_names.index(name):
                self.selection_n = self.roi_names.index(name)
            else:
                self.selection_n = -1
            self.update()

    def _draw_histogram(self):
        self.histogram = self.hist_axis.hist(self.image_data[self.i_index, self.j_index].values.flatten(),
                                             bins=20, density=True)

    def _draw_hist_lines(self):
        self.vlinea = self.hist_axis.axvline(np.nanmin(self.image_data[self.i_index, self.j_index].raw_data))
        self.vlineb = self.hist_axis.axvline(np.nanmax(self.image_data[self.i_index, self.j_index].raw_data))

    def _draw_image(self):
        self.main_image = self.image_axis.imshow(np.squeeze(self.image_data[self.i_index, self.j_index].raw_data),
                                                 extent=self.extent, cmap=self.cmap, origin="lower")
        if self.overlay_data is not None:
            overlay_image = np.full_like(self.overlay_data[self.i_index, self.j_index].raw_data, np.nan)
            overlay_image[:] = self.overlay_data[self.i_index, self.j_index].raw_data
            overlay_image[overlay_image < self.overlay_threshold] = np.nan
            self.overlay_image = self.image_axis.imshow(np.squeeze(overlay_image),
                                                        extent=self.overlay_extent, cmap=self.overlay_cmap,
                                                        origin="lower")

    @property
    def i_index(self):
        try:
            return self.frame.val
        except AttributeError:
            return 0

    @property
    def j_index(self):
        try:
            return self.wavelength.val
        except AttributeError:
            return 0

    def save_roi(self, val):
        if self.drawing:
            z = self.pa_data.get_z_positions()[self.i_index, self.j_index]
            run = self.pa_data.get_run_number()[self.i_index, self.j_index]
            repetition = self.pa_data.get_repetition_numbers()[self.i_index, self.j_index]
            roi_name = self.roi_name
            roi_position = self.roi_position
            roi_output = self.verts
            self.pa_data.add_roi(ROI(roi_output, z, run, repetition, roi_name, roi_position))
            print("Saved ROI", self.roi_name, self.scan_name)
            self.p_selector.set_visible(False)
            self.drawing = False
            self.start_btn.label.set_text("Start ROI")
            self.update()
            self.p_selector = None

    def update_clim(self, val):
        cmin = val[0]
        cmax = val[1]
        self.vlinea.set_data(([cmin, cmin], [0, 1]))
        self.vlineb.set_data(([cmax, cmax], [0, 1]))
        self.main_image.set_clim([cmin, cmax])

    def poly_onselect(self, v):
        self.verts = v

    def start_drawing(self, v):
        if not self.drawing:
            self.selection_n = -1
            self.drawing = True
            self.p_selector = PolygonSelector(self.image_axis, self.poly_onselect)
            self.start_btn.label.set_text("Cancel ROI")
            self.update()
        else:
            self.selection_n = -1
            self.drawing = False
            self.p_selector.set_visible(False)
            self.p_selector = None
            self.start_btn.label.set_text("Start ROI")
            self.update()

    def edit_roi_name(self, _=None):
        if self.selection_n == -1:
            return
        name = self.roi_names[self.selection_n]
        self.pa_data.rename_roi(name, self.roi_name, self.roi_position)
        self.selection_n = -1
        self.update()

    def delete_roi(self, _=None):
        name = self.roi_names[self.selection_n]
        print("Deleting", name)
        self.pa_data.delete_rois(*name.split("/"))
        self.selection_n = -1
        self.update()

    def on_close(self, _=None):
        self.pa_data.close()

    def on_key_press(self, event):
        if event.key == "up":
            self.selection_n += 1
            self.selection_n %= len(self.roi_lines)
        elif event.key == "down":
            self.selection_n -= 1
            if self.selection_n < -1:
                self.selection_n = -1
        if event.key == "left":
            self.frame.set_val(max(self.frame.val - 1, self.frame.valmin))
        elif event.key == "right":
            self.frame.set_val(min(self.frame.val + 1, self.frame.valmax))
        if event.key == "r":
            self.edit_roi_name()
        self.update()

    def __init__(self, pa_data, image_data: ImageSequence, image_extent, overlay=None,
                 overlay_threshold=None, overlay_extent=None,
                 roi_name="unnamed", roi_position="", interpolation=False,
                 draw_all_rois=False, start_wl=0):
        self.draw_all_rois = draw_all_rois
        clinical = pa_data.is_clinical()
        self.frame_type = "z" if not clinical else "repetition"
        self.match_frames = pa_data.get_z_positions() if not clinical else pa_data.get_repetition_numbers()
        self.pa_data = pa_data
        self.interpolation = interpolation

        self.image_data = image_data
        self.overlay_data = overlay
        self.overlay_threshold = overlay_threshold
        self.overlay_extent = overlay_extent or image_extent
        self.overlay_cmap = None
        if overlay is not None:
            self.overlay_cmap = overlay.cmap
        self.fig = plt.figure()
        self.image_axis = plt.subplot2grid((2, 2), (0, 0), 2, fig=self.fig)
        self.hist_axis = plt.subplot2grid((2, 2), (0, 1), fig=self.fig)
        self.hist_axis.set_ylabel("Fraction")
        self.hist_axis.set_xlabel("PA Intensity")
        plt.tight_layout(pad=3)
        plt.subplots_adjust(bottom=0.3)
        self.extent = image_extent
        self.cmap = image_data.cmap
        self.main_image = None
        self.overlay_image = None
        self.histogram = None

        # Draw the main images
        self._draw_image()

        self._draw_histogram()
        self.vlinea = None
        self.vlineb = None
        self._draw_hist_lines()
        # Define the slider and button axes
        self.ax_clims = plt.axes([0.25, 0.15, 0.5, 0.03])
        self.ax_slide = plt.axes([0.25, 0.11, 0.5, 0.03])
        self.ax_frame = plt.axes([0.25, 0.07, 0.5, 0.03])
        self.ax_button = plt.axes([0.25, 0.01, 0.2, 0.05])
        self.ax_start = plt.axes([0.55, 0.01, 0.2, 0.05])

        if self.image_data.shape[1] != 1:
            self.wavelength = Slider(self.ax_slide, "Wavelength", 0,
                                     self.image_data.shape[1] - 1, valinit=start_wl, valstep=1)
        else:
            Constant = namedtuple("ConstantWidget", ["val", "on_changed"])
            self.wavelength = Constant(val=0, on_changed=lambda x: None)
            self.ax_slide.set_visible(False)

        self.frame = Slider(self.ax_frame, "Frame", 0, self.image_data.shape[0] - 1, valinit=0, valstep=1)
        self.previous_frame_number = 0

        self.roi_lines = []
        self.roi_names = []
        self.image_axis.set_prop_cycle(None)
        self.legend = []
        self.legend_y_0 = 0.55
        self.legend_dy = -0.04
        self.legend_x_0 = 0.55
        self.legend_y = 0
        self.selection_n = -1
        self.drawing = False

        self.ax_edit = plt.axes([self.legend_x_0 + 0.2, self.legend_y_0 - 0.1, 0.2, 0.05])
        self.edit_btn = Button(self.ax_edit, "No ROI Selected")
        self.edit_btn.set_active(False)

        self.ax_delete = plt.axes([self.legend_x_0 + 0.2, self.legend_y_0 - 0.3, 0.2, 0.05])
        self.delete_btn = Button(self.ax_delete, "No ROI Selected")
        self.delete_btn.set_active(False)

        self.save_btn = Button(self.ax_button, "Save ROI")
        self.start_btn = Button(self.ax_start, "Start ROI")

        self.clims = RangeSlider(self.ax_clims, "Clim Range",
                                 np.nanmin(self.image_data[self.i_index, self.j_index].raw_data),
                                 np.nanmax(self.image_data[self.i_index, self.j_index].raw_data))

        self.clims.set_min(np.nanmin(self.main_image.get_array()))
        self.clims.set_max(np.nanmax(self.main_image.get_array()))
        self.clims.set_val((np.nanmin(self.main_image.get_array()), np.nanmax(self.main_image.get_array())))

        self.ax_text = plt.axes([0.3, 0.95, 0.4, 0.04])
        self.ax_text.axis("off")
        self.scan_name = pa_data.get_scan_name()
        self.roi_name = roi_name
        self.roi_position = roi_position
        self.ax_text.annotate(self.scan_name + ": " + self.roi_name + " " + self.roi_position, (0.5, 0),
                              xycoords="axes fraction", annotation_clip=False,
                              horizontalalignment="center", verticalalignment="bottom")

        Constant = namedtuple("ConstantWidget", ["set_text"])
        self.z_text = Constant(lambda x: None)
        z = pa_data.get_z_positions()[self.i_index, self.j_index]
        if not np.isnan(z):
            self.z_text = self.fig.text(0.6, 0.19, f"z = {z:.2f} mm")

        wl = pa_data.get_wavelengths()[self.j_index]
        self.wl_text = self.fig.text(0.3, 0.19, f"wavelength = {wl:.0f} nm")

        self.p_selector = None
        self.verts = None
        self.fig.canvas.mpl_connect('pick_event', self.on_pick)
        self.update()
        self.update_clim(self.main_image.get_clim())

        self.frame.on_changed(self.update)
        self.wavelength.on_changed(self.update)
        self.save_btn.on_clicked(self.save_roi)
        self.start_btn.on_clicked(self.start_drawing)
        self.clims.on_changed(self.update_clim)
        self.edit_btn.on_clicked(self.edit_roi_name)
        self.delete_btn.on_clicked(self.delete_roi)
        self.fig.canvas.mpl_connect('close_event', self.on_close)
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
        # fm = self.fig.canvas.manager.full_screen_toggle()
        plt.show()


def main():
    p = init_argparse()
    args = p.parse_args()

    DATA_FOLDER = args.input

    if os.path.isfile(DATA_FOLDER):
        data_files = [DATA_FOLDER]
    else:
        data_files = sorted(glob.glob(join(DATA_FOLDER, "**", "*.hdf5"), recursive=True), key=sort_key)

    print(args.input)
    for file in data_files:
        if args.filter is not None:
            if split(file)[-1] != "Scan_" + args.filter + ".hdf5":
                continue
        if args.start_scan_number is not None:
            file_name = split(file)[-1]
            if "_" in file_name:
                number = int(file_name.split("_")[1].split(".")[0])
                if number < args.start_scan_number:
                    continue

        data = PAData.from_hdf5(file, "r+")
        wl_i = np.argmin(np.abs(data.get_wavelengths() - args.wavelength))

        if args.filtername is not None:
            if str.lower(args.filtername) not in str.lower(data.get_scan_name()):
                continue
            print(file)

        if args.clear:
            if args.name == "unnamed" and args.position == "":
                data.delete_rois()
            else:
                data.delete_rois(args.name + "_" + args.position)
            continue

        if not data.get_recon_types():
            print(f"Skipping {file} because no reconstructions are present.")
            continue

        print(split(file)[-1])

        methods = data.get_recon_types()
        print(methods)
        method = sorted(methods)[args.recon]

        if len(methods) > 1:
            print(f"Multiple reconstructions available, using {method}.")

        recon = data.get_scan_reconstructions()[method]
        
        # Load to memory:
        if args.loadall:
            print("Loading all")
            recon.raw_data = recon.raw_data[:]
        
        extents = recon.extent

        thb = data.get_scan_thb()

        if args.drawthb:
            recon = thb[(method, "0")]

        ROIDrawer(data, recon, extents, roi_name=args.name, roi_position=args.position,
                                 interpolation=args.interpolation, draw_all_rois=args.drawallrois, start_wl = wl_i)


if __name__ == "__main__":
    main()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from typing import Sequence

import numpy as np
import pylops
from pylops.optimization.leastsquares import regularized_inversion
from pylops.signalprocessing import Convolve1D, Convolve2D
from scipy.sparse.linalg import LinearOperator as CPULinOp

import warnings

from .cuda_implementation import get_model as get_model_gpu_single_c
from .cuda_implementation_refraction import get_model as get_model_gpu_double_c
from .numpy_implementation import get_model as get_model_cpu_single_c
from .. import ReconstructionAlgorithm
from ...core.image_structures.pa_time_data import PATimeSeries
from ... import PAData

try:
    cuda_enabled = True
    from cupyx.scipy.sparse.linalg import LinearOperator as GPULinOp
    import cupy as cp
except:
    cuda_enabled = False


class ModelBasedReconstruction(ReconstructionAlgorithm):
    """Model based reconstruction algorithm processor.
    """
    def generate_model(self, detx, dety, fs, dx, nx, x_0, nt, gpu=True, cache=False, **kwargs):
        """

        Parameters
        ----------
        detx
        dety
        fs
        dx
        nx
        x_0
        nt
        gpu
        cache
        kwargs

        Returns
        -------

        """
        if gpu:
            get_model_single_c = get_model_gpu_single_c
            get_model_double_c = get_model_gpu_double_c
            LinearOperator = GPULinOp
        else:
            get_model_double_c = get_model_cpu_single_c
            get_model_single_c = get_model_cpu_single_c
            LinearOperator = CPULinOp
        model_type = self.custom_params.get("model_type", "single_sos")
        regulariser = self.custom_params.get("regulariser", self.custom_params.get("regularizer", None))
        if regulariser not in ["identity", "laplacian", "TV"]:
            raise ValueError("Invalid regulariser.")
        reg_lambda = self.custom_params.get("reg_lambda", 1)
        iter_lim = self.custom_params.get("iter_lim", 10)

        # generate a model.
        if model_type == "single_sos":
            c = kwargs.get("c", None) or self.custom_params.get("c")
            dl = c / fs
            model_matrix = get_model_single_c(detx, dety, dl, dx, nx,
                                              x_0, nt, cache=cache)
        elif model_type == "refraction":
            c0 = kwargs.get("c0", None) or self.custom_params.get("c0")
            c1 = kwargs.get("c1", None) or self.custom_params.get("c1")
            y_cutoff = kwargs.get("y_cutoff", None) or self.custom_params.get("y_cutoff")
            model_matrix = get_model_double_c(detx, dety, c0 / fs, c1 / fs,
                                              y_cutoff, dx, nx, x_0, nt, cache=cache)
        else:
            raise NotImplementedError(f"Model {model_type} is unavailable.")
        self._raw_model = model_matrix
        # Force dodgy duck typing to enable product with irf convolution.
        model_matrix.matvec = lambda x: None

        irf = kwargs.get("irf_model", None)
        if irf is None:
            irf = self.custom_params.get("irf_model", None)

        ndet = detx.shape[0]
        if irf is not None:
            convolve_irf = Convolve1D((ndet, nt), irf, axis=1, offset=nt // 2,
                                      dtype=np.float32)

            def matvec(x):
                return convolve_irf @ (model_matrix @ x)

            def rmatvec(x):
                z = convolve_irf.H @ x
                return model_matrix.H @ z

            full_model = LinearOperator(model_matrix.shape, matvec=matvec, rmatvec=rmatvec, dtype=np.float32)
            full_model.data = 1
        else:
            def matvec(x):
                return model_matrix @ x

            def rmatvec(x):
                return model_matrix.H @ x
            full_model = LinearOperator(model_matrix.shape, matvec=matvec, rmatvec=rmatvec, dtype=np.float32)
            full_model.data = 1
        inv_args = {}

        if regulariser == "identity":
            reg = pylops.Identity(nx * nx)
        elif regulariser == "laplacian":
            if gpu:
                reg_mat = cp.array([[-1, -1, -1], [-1, 9, -1], [-1, -1, -1]])
            else:
                reg_mat = np.array([[-1, -1, -1], [-1, 9, -1], [-1, -1, -1]])
            reg = Convolve2D([nx, nx], reg_mat)
        elif regulariser == "TV":
            reg = [
                pylops.FirstDerivative(dims=(nx, nx), axis=0, edge=True, kind="backward", dtype=np.float32
                                       ),
                pylops.FirstDerivative(dims=(nx, nx), axis=1, edge=True, kind="backward", dtype=np.float32
                                       ),
            ]

            inv_args["RegsL2"] = [pylops.Identity(nx * nx)]
            inv_args["epsRL2s"] = [reg_lambda[1]]
            inv_args["mu"] = kwargs.get("mu", 1.)
            inv_args["tol"] = kwargs.get("tol", 1e-10)
            inv_args["tau"] = kwargs.get("tau", 1.)
        else:
            raise NotImplementedError(f"Regulariser {regulariser} is not available.")

        self._model_matrix = full_model, reg
        if regulariser in ["identity", "laplacian"]:
            if gpu:
                inv_args["niter"] = iter_lim
                inv_args["tol"] = 1e-6
            else:
                inv_args["iter_lim"] = iter_lim
            return lambda y: regularized_inversion(full_model, y,
                                                   [reg], epsRs=[reg_lambda], show=False, **inv_args)[0]
        else:
            self._full_model = full_model
            self._reg = reg
            self._niter_outer = iter_lim[0]
            self._niter_inner = iter_lim[1]
            self._epsRL1s = [reg_lambda[0], reg_lambda[0]]
            self._inv_args = inv_args
            return lambda y: pylops.optimization.sparsity.splitbregman(
                full_model,
                y,
                reg,
                niter_outer=iter_lim[0],
                niter_inner=iter_lim[1],
                epsRL1s=[reg_lambda[0], reg_lambda[0]],
                show=False, **inv_args
            )[0]

    def __init__(self,
                 n_pixels: Sequence[int],
                 field_of_view: Sequence[float],
                 kwargs_model=None,
                 pa_example: "PAData"=None,
                 **kwargs):
        if kwargs.get("regulariser", None) is None:
            kwargs["regulariser"] = "identity"
        super().__init__(n_pixels, field_of_view, **kwargs)
        self._model_matrix = None
        self._raw_model = None
        if kwargs_model is None:
            kwargs_model = {}

        if pa_example is not None:
            kwargs["geometry"] = pa_example.get_scan_geometry()
            kwargs["fs_model"] = pa_example.get_sampling_frequency()
            kwargs["nt"] = pa_example.get_time_series().shape[-1]
            kwargs_model["c"] = pa_example.get_speed_of_sound()
            kwargs_model["irf_model"] = pa_example.get_impulse_response()[:]

        # Note that super init puts kwargs in self.custom_params
        if "geometry" in kwargs:
            # identify x and y:
            detectors = kwargs["geometry"]
            indices = []
            for i in range(3):
                if np.all(kwargs["geometry"][:, i] != 0.):
                    indices.append(i)
            self._indices = indices

            detx = detectors[:, indices[0]]
            dety = detectors[:, indices[1]]

            fs = kwargs["fs_model"]
            self._fs = fs
            nx = n_pixels[indices[0]]
            self._nx = nx
            dx = field_of_view[indices[0]] / (nx - 1)
            self._dx = dx
            x_0 = -field_of_view[indices[0]] / 2
            self._x_0 = x_0
            nt = kwargs["nt"]

            gpu = kwargs.get("gpu", cuda_enabled)
            cache = kwargs.get("cache", False)

            self._model = self.generate_model(detx, dety, fs, dx, nx, x_0, nt, gpu=gpu, cache=cache, **kwargs_model)
        else:
            self._fs = None
            self._nx = None
            self._dx = None
            self._x_0 = None
            self._model = None
            self._indices = None
        warnings.warn("This class is experimental. If you would like to contribute to further development of this "
                      "approach, please get in touch.")

    def pre_prepare_data(self, x: PATimeSeries):
        """

        Parameters
        ----------
        x
        """
        # really just for speed of sound setting
        pass

    def reconstruct(self, signal: np.ndarray, fs: float,
                    geometry: np.ndarray,
                    n_pixels: Sequence[int],
                    field_of_view: Sequence[float],
                    speed_of_sound=None,
                    **kwargs) -> np.ndarray:
        """

        Parameters
        ----------
        signal
        fs
        geometry
        n_pixels
        field_of_view
        speed_of_sound
        kwargs

        Returns
        -------

        """
        gpu = self.custom_params.get("gpu", cuda_enabled)
        if self._model is not None:
            model = self._model
            fs = self._fs
            nx = self._nx
            dx = self._dx
            x_0 = self._x_0
            indices = self._indices
        else:
            detectors = geometry
            indices = []
            for i in range(3):
                if np.all(kwargs["geometry"][:, i] != 0.):
                    indices.append(i)

            assert len(indices) == 2
            detx = detectors[:, indices[0]]
            dety = detectors[:, indices[1]]

            do_irf = self.custom_params.get("do_irf", True)
            irf = kwargs["irf"] if do_irf else None

            nx = n_pixels[indices[0]]
            dx = field_of_view[indices[0]] / (nx - 1)
            x_0 = -field_of_view[indices[0]] / 2
            nt = signal.shape[-1]

            model = self.generate_model(detx, dety, fs, dx, nx, x_0, nt, irf_model=irf,
                                        gpu=gpu)
        if gpu:
            t = cp.array(signal)
        else:
            t = np.array(signal)

        or_shape = t.shape

        output_shape = [1, 1, 1]
        for i in indices:
            output_shape[i] = nx
        # Reverse output shape to be (z, y, x) order
        output_shape = tuple(output_shape[::-1])

        t = t.reshape((np.product(t.shape[:-2]),) + t.shape[-2:])
        output = np.zeros((t.shape[0],) + output_shape)

        for i, t0 in enumerate(t):
            result = model(t0.flatten()).reshape(output_shape)
            if gpu:
                output[i] = result.get()
            else:
                output[i] = result

        return output.reshape(or_shape[:-2] + output.shape[-3:])

    @staticmethod
    def get_algorithm_name() -> str:
        """

        Returns
        -------

        """
        return "Model Based Reconstruction"



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import unittest
from os.path import join

from patato import PreProcessor
from patato.core.image_structures.reconstruction_image import Reconstruction
from patato.core.image_structures.unmixed_image import UnmixedData
from patato.data.get_example_datasets import get_patato_data_folder
from patato.io.msot_data import PAData
from patato.unmixing.spectra import OxyHaemoglobin, Haemoglobin
from patato.recon import ReferenceBackprojection
from patato.unmixing.unmixer import SpectralUnmixer
from patato.utils.pipeline import run_pipeline

from make_dummy_dataset import make_dummy_dataset


class TestPipelines(unittest.TestCase):
    def setUp(self) -> None:
        make_dummy_dataset()

    def test_full_run(self):
        data_folder = join(get_patato_data_folder(), "test")
        dummy_dataset = join(data_folder, "Scan_1.hdf5")

        pa = PAData.from_hdf5(dummy_dataset, "r+")[:, 0:2]

        preproc = PreProcessor(time_factor=3, detector_factor=2)
        reconstructor = ReferenceBackprojection([333, 1, 333], [0.025, 1, 0.025])
        unmixer = SpectralUnmixer([OxyHaemoglobin(), Haemoglobin()], pa.get_wavelengths())

        preproc.add_child(reconstructor)
        reconstructor.add_child(unmixer)

        results, additional_results = run_pipeline(preproc, pa.get_time_series(), pa, n_batch=-1, save_results=False)

        self.assertEqual(len(additional_results), 0)
        self.assertEqual(type(results[0]), Reconstruction)
        self.assertEqual(type(results[1]), UnmixedData)


#  Copyright (c) Thomas Else 2023.
#  License: MIT

from typing import Sequence

import numpy as np

from .backprojection_implementation.jax_implementation import full_recon
from .reconstruction_algorithm import ReconstructionAlgorithm

try:
    import jax
except ImportError:
    print("""WARNING: JAX must be installed to support the standard backprojection algorithm and filtering.
    Alternatively, try the Numpy implementations.
    """)
    jax = None


class ReferenceBackprojection(ReconstructionAlgorithm):
    """
    Reference backprojection: Uses JAX in the background.
    """

    def reconstruct(self, time_series: np.ndarray,
                    fs: float,
                    geometry: np.ndarray, n_pixels: Sequence[int],
                    field_of_view: Sequence[float],
                    speed_of_sound,
                    **kwargs) -> np.ndarray:
        """

        Parameters
        ----------
        time_series
        fs
        geometry
        n_pixels
        field_of_view
        speed_of_sound
        kwargs

        Returns
        -------

        """

        # Get parameters:
        dl = speed_of_sound / fs

        # Reshape frames so that we can loop through to reconstruct
        original_shape = time_series.shape[:-2]
        frames = int(np.product(original_shape))
        signal = time_series.reshape((frames,) + time_series.shape[-2:])
        dx = field_of_view[0] / (n_pixels[0] - 1) if n_pixels[0] != 1 else 0
        dy = field_of_view[1] / (n_pixels[1] - 1) if n_pixels[1] != 1 else 0
        dz = field_of_view[2] / (n_pixels[2] - 1) if n_pixels[2] != 1 else 0

        recon_all = jax.vmap(full_recon, in_axes=(0,) + (None,) * 8, out_axes=0)

        output = recon_all(signal, geometry, dl, n_pixels[0], n_pixels[1], n_pixels[2],
                           dx, dy, dz)

        return output.reshape(original_shape + tuple(n_pixels)[::-1])

    @staticmethod
    def get_algorithm_name() -> str:
        """

        Returns
        -------

        """
        return "Reference Backprojection"



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from __future__ import annotations

from typing import Union, TYPE_CHECKING, Tuple, Optional

from .processing_algorithm import ProcessingResult, TimeSeriesProcessingAlgorithm

if TYPE_CHECKING:
    from ..io.msot_data import PAData

from ..core.image_structures.pa_time_data import PATimeSeries

try:
    import cupy as cp
    from cupy.fft import fft, ifft, fftshift, fftfreq
except ImportError:
    cp = None

from ..io.attribute_tags import PreprocessingAttributeTags
import time
from patato.unmixing.spectra import SPECTRA_NAMES
import logging


class GPUMSOTPreProcessor(TimeSeriesProcessingAlgorithm):
    @staticmethod
    def get_algorithm_name() -> str:
        return "GPU Standard Preprocessor"

    @staticmethod
    def get_hdf5_group_name() -> Union[str, None]:
        return None

    def __init__(self, time_factor=3, detector_factor=2,
                 irf=True, hilbert=True, lp_filter=None,
                 hp_filter=None, filter_window_size=512,
                 window: Union[str, None] = "hann", absolute: Union[bool, str] = None,
                 couplant_correction=None, couplant_path_length=0):
        super().__init__()
        self.time_factor = time_factor
        self.detector_factor = detector_factor
        absolute = "imag" if absolute is None and hilbert else absolute
        self.irf_correct = irf
        self.hilbert = hilbert
        self.lp_filter = lp_filter
        self.hp_filter = hp_filter
        self.n_filter = filter_window_size
        self.window = window
        self.absolute = absolute

        if couplant_correction is not None:
            self.couplant_correction = SPECTRA_NAMES[couplant_correction]
        else:
            self.couplant_correction = None
        self.couplant_path_length = couplant_path_length

    def run(self, time_series: PATimeSeries,
            pa_data: PAData, irf=None, detectors=None, **kwargs) -> Tuple["PATimeSeries", dict,
                                                                          Optional[ProcessingResult]]:
        if irf is None:
            irf = pa_data.get_impulse_response()
        if detectors is None:
            detectors = pa_data.get_scan_geometry()
        fs = time_series.attributes["fs"]

        if pa_data.get_overall_correction_factor() is not None:
            overall_correction_factor = cp.array(pa_data.get_overall_correction_factor())

        # Generate the filter
        ft_filter = self.make_filter(time_series.shape[-1],
                                     fs,
                                     cp.array(irf) if self.irf_correct else None,
                                     self.hilbert, self.lp_filter,
                                     self.hp_filter,
                                     n_filter=self.n_filter,
                                     window=self.window)

        new_time_series, new_parameters = self._run(time_series,
                                                    ft_filter, cp.array(detectors),
                                                    cp.array(overall_correction_factor), **kwargs)

        # Update the results' attributes.
        for a in time_series.attributes:
            if a not in new_time_series.attributes:
                new_time_series.attributes[a] = time_series.attributes[a]

        new_time_series.attributes[PreprocessingAttributeTags.IMPULSE_RESPONSE] = self.irf_correct
        new_time_series.attributes[PreprocessingAttributeTags.PROCESSING_ALGORITHM] = self.get_algorithm_name()
        new_time_series.attributes[PreprocessingAttributeTags.WINDOW_SIZE] = self.window
        new_time_series.attributes[PreprocessingAttributeTags.ENVELOPE_DETECTION] = self.absolute == "abs"
        new_time_series.attributes[PreprocessingAttributeTags.HILBERT_TRANSFORM] = self.hilbert
        new_time_series.attributes[PreprocessingAttributeTags.DETECTOR_INTERPOLATION] = self.detector_factor
        new_time_series.attributes[PreprocessingAttributeTags.TIME_INTERPOLATION] = self.time_factor
        new_time_series.attributes[PreprocessingAttributeTags.LOW_PASS_FILTER] = self.lp_filter
        new_time_series.attributes[PreprocessingAttributeTags.HIGH_PASS_FILTER] = self.hp_filter
        new_time_series.attributes["CorrectionFactorApplied"] = overall_correction_factor is not None

        # TODO: replace pa_data with new attributes for further processing steps.

        return new_time_series, new_parameters, None

    def _run(self, time_series: PATimeSeries, ft_filter, detectors, overall_correction_factor, **kwargs) -> Tuple[
        PATimeSeries, dict]:
        new_parameters = {}

        # Subtract mean
        extend = (slice(None, None),) * (time_series.raw_data.ndim - 1) + (None,)

        # Subtract mean
        t = time.time()
        raw_data = cp.array(time_series.raw_data)
        time_series.raw_data = raw_data - cp.mean(raw_data, axis=-1)[extend]
        logging.debug(f"Mean subtraction took {time.time() - t}s")
        # Apply a fourier domain filter.

        t = time.time()
        time_series_ft = fft(time_series.raw_data, axis=-1)
        time_series_ft = self.apply_filter(time_series_ft, ft_filter=ft_filter)

        operation = cp.real if self.absolute == "real" or self.absolute is None else cp.imag if self.absolute == "imag" else cp.abs
        time_series.raw_data = operation(ifft(time_series_ft, axis=-1))

        # Go back to the time domain.
        logging.debug(f"Filter took {time.time() - t}s")

        # Apply interpolation in time and detector domains.
        time_series, interp_params = self.interpolate(time_series, detectors)
        new_parameters.update(interp_params)

        # Apply energy correction factor
        extend = (slice(None, None),) * overall_correction_factor.ndim + (None, None)
        time_series.raw_data /= overall_correction_factor[extend]

        # Apply the couplant correction.
        if self.couplant_correction is not None:
            # TODO: This needs rewriting.
            raise NotImplementedError("Couplant correction is not implemented.")
            # extend = (None,) * (time_series.ndim - 3) + (slice(None, None), None, None)
            # time_series.raw_data *= \
            #     np.exp(self.couplant_path_length * self.couplant_correction.get_spectrum(pa_data.get_wavelengths()))[
            #         extend]

        time_series.raw_data = time_series.raw_data.get()
        return time_series, new_parameters

    def interpolate(self, time_series: PATimeSeries, detectors, exact_ratios=True) -> Tuple[PATimeSeries, dict]:
        # Interpolate the data in the time and detector domains.
        # Interpolate in the detector domain
        if self.time_factor == 1 and self.detector_factor == 1:
            return time_series, {"geometry": detectors}

        signal = time_series.raw_data

        # Get the new indices on which to interpolate.
        detector_ind = cp.arange(detectors.shape[0])
        new_detector_ind = cp.arange((detectors.shape[0] - 1) * self.detector_factor + 1) / self.detector_factor
        if exact_ratios:
            new_detector_ind = cp.linspace(0, detectors.shape[0] - 1, self.detector_factor * detectors.shape[0])

        # Interpolate in the sample domain
        cp.arange(signal.shape[-1])
        new_samp_ind = cp.arange((signal.shape[-1] - 1) * self.time_factor + 1) / self.time_factor
        if exact_ratios:
            new_samp_ind = cp.linspace(0, signal.shape[-1] - 1, self.time_factor * signal.shape[-1])

        new_signal = cp.zeros(signal.shape[:-2] + (new_detector_ind.shape[0], new_samp_ind.shape[0]),
                              dtype=signal.dtype)
        t = time.time()

        sind, dind = cp.meshgrid(new_samp_ind, new_detector_ind)

        # print(dind.shape, signal.shape)
        new_signal += signal[..., dind.astype(cp.int32), sind.astype(cp.int32)] * (1 - (dind-cp.floor(dind))) * (1 - (sind-cp.floor(sind)))
        new_signal += signal[..., dind.astype(cp.int32) + 1, sind.astype(cp.int32)] * (dind-cp.floor(dind)) * (1 - (sind-cp.floor(sind)))
        new_signal += signal[..., dind.astype(cp.int32), sind.astype(cp.int32) + 1] * (1 - (dind-cp.floor(dind))) * (sind-cp.floor(sind))
        new_signal += signal[..., dind.astype(cp.int32) + 1, sind.astype(cp.int32) + 1] * (dind-cp.floor(dind)) * (sind-cp.floor(sind))
        #
        # assert 1==2

        detectors = cp.apply_along_axis(lambda x: cp.interp(new_detector_ind, detector_ind, x),
                                        0, detectors)

        logging.debug(f"Interpolation took {time.time() - t}.")

        # Update the xarray coordinates of the new dataset.
        coords = dict(time_series.da.coords)
        coords["detectors"] = new_detector_ind.get()
        coords["timeseries"] = new_samp_ind.get()

        attributes = dict(time_series.da.attrs)
        attributes["fs"] *= self.time_factor

        new_data = PATimeSeries(new_signal, time_series.da.dims, coords, attributes=attributes)
        return new_data, {"geometry": detectors.get()}

    @staticmethod
    def apply_filter(pa_data: cp.ndarray, ft_filter) -> cp.ndarray:
        extend = (None,) * (pa_data.ndim - 1) + (slice(None, None),)
        pa_data *= ft_filter[extend]
        return pa_data

    @staticmethod
    def make_filter(n_samples, fs, irf,
                    hilbert, lp_filter,
                    hp_filter, rise=0.2,
                    n_filter=1024, window=None) -> cp.ndarray:
        # at the moment, it looks like it is shifting the data a bit??
        # Impulse Response Correction
        output = cp.ones((n_samples,), dtype=cp.cdouble)
        if irf is not None:
            irf_shifted = cp.zeros_like(irf)
            irf_shifted[:irf.shape[0] // 2] = irf[irf.shape[0] // 2:]
            irf_shifted[-irf.shape[0] // 2:] = irf[:irf.shape[0] // 2]
            output *= cp.conj(fft(irf_shifted)) / cp.abs(fft(irf_shifted)) ** 2
            from scipy.signal.windows import hann
            output *= fftshift(hann(n_samples))

        # Hilbert Transform
        frequencies = fftfreq(n_samples)
        if hilbert:
            output *= (1 + cp.sign(frequencies)) / 2

        frequencies = cp.abs(fftfreq(n_filter, 1 / fs))
        fir_filter = cp.ones_like(frequencies, dtype=cp.cdouble)
        if hp_filter is not None:
            fir_filter[frequencies < hp_filter * (1 - rise)] = 0
            in_rise = cp.logical_and(frequencies > hp_filter * (1 - rise), frequencies < hp_filter)
            fir_filter[in_rise] = (frequencies[in_rise] - hp_filter * (1 - rise)) / (hp_filter * rise)
        if lp_filter is not None:
            fir_filter[frequencies > lp_filter * (1 + rise)] = 0
            in_rise = cp.logical_and(frequencies < lp_filter * (1 + rise), frequencies > lp_filter)
            fir_filter[in_rise] = 1 - (frequencies[in_rise] - lp_filter) / (lp_filter * rise)

        time_series = ifft(fir_filter)

        if window == "hann":
            from scipy.signal.windows import hann
            time_series *= fftshift(hann(n_filter))

        filter_time = cp.zeros_like(output)
        filter_time[:n_filter // 2] = time_series[:n_filter // 2]
        filter_time[-n_filter // 2:] = time_series[-n_filter // 2:]
        fir_filter = fft(filter_time)
        output *= fir_filter
        return output



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from typing import Sequence

from .gpu_preprocessing_algorithm import GPUMSOTPreProcessor
from .preprocessing_algorithm import NumpyPreProcessor
from .jax_preprocessing_algorithm import PreProcessor
from .processing_algorithm import TimeSeriesProcessingAlgorithm


PREPROCESSING_METHODS: Sequence[type(TimeSeriesProcessingAlgorithm)] = [NumpyPreProcessor,
                                                                        GPUMSOTPreProcessor,
                                                                        PreProcessor]

PREPROCESSING_NAMES = {x.get_algorithm_name(): x for x in PREPROCESSING_METHODS}



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
Time-series data processing module
=========================================

This module provides functions for commonly-used pre-processing procedures used in photoacoustic imaging.
This includes, for example, band-pass filtering, Hilbert transforms and interpolation.
"""


"""
PA raw data implements the data structure for photoacoustic time series data.
"""

#  Copyright (c) Thomas Else 2023.
#  License: MIT

from abc import ABC
from typing import TYPE_CHECKING


from ...core.image_structures.image_sequence import DataSequence
from ...io.attribute_tags import HDF5Tags

if TYPE_CHECKING:
    pass


class PARawData(DataSequence, ABC):
    """
    PARawData is the abstract data structure for time series data. This can be in the pure time series format or the
    fourier way.
    """

    def __add__(self, other):
        raise NotImplementedError()
    n_im_dim = 2

    def __init__(self, data, dimensions, coordinates=None, attributes=None, hdf5_sub_name=None,
                 algorithm_id=None):
        """

        Parameters
        ----------
        data
        dimensions
        coordinates
        attributes
        hdf5_sub_name
        algorithm_id
        """
        super().__init__(data, dimensions, coordinates, attributes,
                         hdf5_sub_name=hdf5_sub_name, algorithm_id=algorithm_id)
        self.save_output = False

    @staticmethod
    def is_single_instance():
        """

        Returns
        -------

        """
        return True

    @staticmethod
    def get_ax1_label_meaning():
        """

        Returns
        -------

        """
        return HDF5Tags.WAVELENGTH



#  Copyright (c) Thomas Else 2023.
#  License: MIT


import numpy as np


def get_basic_p0(nx=333, dx=75e-6, r=0.005):
    """

    Get a basic p0 for the simulation.
    Parameters
    ----------
    nx : int
    dx : float

    Returns
    -------
    p0 : ndarray

    """

    x, y = np.ogrid[:nx, :nx]
    x = x.astype(np.float64) - (nx - 1) / 2
    y = y.astype(np.float64) - (nx - 1) / 2
    x = x * dx
    y = y * dx
    r2 = r ** 2
    p0 = r2 - (x ** 2 + y ** 2)
    p0[p0 < 0] = 0
    return p0


def generate_basic_model(ndet=256):
    """
    Generate a basic forward model for a photoacoustic system.

    Parameters
    ----------
    ndet

    Returns
    -------
    tuple of array_like, array_like
    """
    try:
        import cupy as cp
        from ..recon.model_based.cuda_implementation import get_model
        CUDA_ENABLED = True
    except ImportError:
        from ..recon.model_based.numpy_implementation import get_model
        CUDA_ENABLED = False

    theta = np.linspace(0, 2 * np.pi, ndet, endpoint=False)
    detectors = np.array([np.cos(theta), np.sin(theta), np.zeros_like(theta)]).T * 0.04

    model = get_model(detectors[:, 0], detectors[:, 1], 1500 / 4e7, 75e-6, 333, -0.0125, 2030, cache=False)
    if CUDA_ENABLED:
        model = model.get()
    model.eliminate_zeros()
    model.sort_indices()
    return model, detectors


def generate_basic_simulation(ndet=256):
    """
    Generate a simulation of the photoacoustic time series for a given number of detectors in a circular geometry.
    The initial pressure distribution used is a truncated paraboloid, which has an analytical solution for the
    time series data. Here, we use the forward photacoustic model from the model-based reconstruction code.

    Parameters
    ----------
    ndet: int - Number of detectors

    Returns
    -------
    tuple of NDArray, NDArray, NDArray

    """
    model, detectors = generate_basic_model(ndet)
    p0 = get_basic_p0()
    ts = (model @ p0.flatten()).reshape(ndet, 2030)
    return p0, ts, detectors



import numpy as np
import pacfish as pf
from pacfish import MetaDatum


class PatHDF5AdapterToIPASCFormat(pf.BaseAdapter):
    def generate_device_meta_data(self) -> dict:
        return {}

    def __init__(self, hdf5_file):
        self.hdf5_file = hdf5_file
        super(PatHDF5AdapterToIPASCFormat, self).__init__()

    def generate_binary_data(self) -> np.ndarray:
        # iThera definition: [frames, wavelengths, detectors, time_series]
        # IPASC definition: [detectors, time_series, wavelength, frames]
        time_series = self.hdf5_file["raw_data"]
        time_series = np.swapaxes(time_series, 2, 0)  # [detectors, wavelengths, frames, time_series]
        time_series = np.swapaxes(time_series, 3, 1)  # [detectors, time_series, frames, wavelength]
        time_series = np.swapaxes(time_series, 3, 2)  # [detectors, time_series, wavelength, frames]
        return time_series

    def generate_meta_data_device(self) -> dict:
        device_metadata_creator = pf.DeviceMetaDataCreator()
        for array_element in np.asarray(self.hdf5_file["GEOMETRY"]):
            det_element = pf.DetectionElementCreator()
            det_element.set_detector_position(array_element)
            device_metadata_creator.add_detection_element(det_element.get_dictionary())
        return device_metadata_creator.finalize_device_meta_data()

    def set_metadata_value(self, metadata_tag: MetaDatum) -> object:
        if metadata_tag == pf.MetadataAcquisitionTags.PULSE_ENERGY:
            return np.asarray(self.hdf5_file['OverallCorrectionFactor'])
        if metadata_tag == pf.MetadataAcquisitionTags.TEMPERATURE_CONTROL:
            return np.asarray(self.hdf5_file["TEMPERATURE"])
        if metadata_tag == pf.MetadataAcquisitionTags.MEASUREMENT_SPATIAL_POSES:
            # TODO convert each element into [0, POS, 0, 0, 0, 0]
            orig_shape = self.hdf5_file["Z-POS"]
            poses = []
            for y_pos in np.asarray(self.hdf5_file["Z-POS"]).reshape((-1, )):
                poses.append([0, y_pos, 0, 0, 0])

            return np.asarray(poses).reshape((orig_shape[0], orig_shape[1], -1))
        if metadata_tag == pf.MetadataAcquisitionTags.MEASUREMENT_TIMESTAMPS:
            return np.asarray(self.hdf5_file["timestamp"])
        if metadata_tag == pf.MetadataAcquisitionTags.ACQUISITION_WAVELENGTHS:
            return np.asarray(self.hdf5_file["wavelengths"])
        return None



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import argparse
import glob
from os.path import join, split

import h5py
from ..utils import sort_key
from tabulate import tabulate


def init_argparse():
    parser = argparse.ArgumentParser(description="Look at the status of a data folder.")
    parser.add_argument('input', type=str, help="Data Folder")
    parser.add_argument(
        "-v", "--version", action="version",
        version=f"{parser.prog} version 0.1"
    )
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()

    names = []
    cs = []
    fnames = []
    print(args.input)

    for filename in sorted(glob.glob(join(args.input, "*.hdf5")), key=sort_key):
        data_file = h5py.File(filename, "r")

        # 1. Check raw data and print name

        if "raw_data" in data_file:
            fname = split(filename)[-1]
            name = data_file["raw_data"].attrs["name"]
            sos = str(data_file["raw_data"].attrs.get("speedofsound", "NOT SET"))
            if len(names) >= 1:
                if names[-1].split("-")[0].split("_")[0] != name.split("-")[0].split("_")[0] and names[-1][0] != "=":
                    names.append("==========")
                    cs.append("==========")
                    fnames.append("=========")
            names.append(name)
            fnames.append(fname)
            cs.append(sos)

    print(tabulate(zip(fnames, names, cs), headers=["File Name", "Scan Name", "Speed of Sound"]))



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import glob
import xml.dom.minidom
from os.path import split, join

import numpy as np
from ...core.image_structures.reconstruction_image import Reconstruction
from ..hdf.fileimporter import ReaderInterface
from ..ithera import load_ithera_irf
import warnings

def xml_to_dict(x):
    if x.nodeType == 3:
        return x.nodeValue
    elif not x.childNodes:
        return x.nodeName, x.nodeValue
    else:
        dicts = [xml_to_dict(y) for y in x.childNodes]
        v = [y for y in dicts if type(y) not in [str, list] or y[0] != "\n"]
        # convert to floats/ints:
        v = [y if type(y) is not str else int(y) if y.isnumeric() else float(y) if y.replace(".",
                                                                                             "").isnumeric() else y == "true" if y in [
            "true", "false"] else y for y in v]
        if type(v[0]) == tuple and all([y == v[0][0] for y, _ in v]):
            v = [y for _, y in v]
        elif type(v[0]) == tuple:
            v = dict(v)
        if type(v) == list and len(v) == 1:
            v = v[0]
        return x.nodeName, v


class iTheraMSOT(ReaderInterface):
    """An interface for iThera MSOT datasets.
    """
    def _get_rois(self):
        # In future, can extend this to enable import of iThera ROIS.
        pass

    def _get_segmentation(self):
        return None

    def _get_datasets(self):
        # Only support reconstructions for now. I think this could be extended easily enough to support
        # unmixed etc though
        # could easily get an error here in older versions of msot data format, so will try except - might need to check
        # that this at least doesn't fail on old versions...
        try:
            f = xml_to_dict(self.xml_tree.getElementsByTagName("ReconNodes")[0])
        except IndexError:
            return {}
        reconstructions = []
        rec_list = f[1] if type(f[1]) == list else [f[1]]

        for r in rec_list:
            if r is None:
                continue
            guid = r["GUID"]
            file = join(self.scan_folder, "RECONs", guid + ".bin")
            ns = [r["FIELD-OF-VIEW"]["PixelCount"][a] for a in "XYZ"]
            try:
                recon = np.memmap(file, dtype=np.single)[:self.nframes * self.nwavelengths * ns[0] * ns[1] * ns[2]].reshape(
                   (self.nframes,
                    self.nwavelengths,) + tuple(ns))
            except ValueError:
                warnings.warn("Warning: unable to import iThera reconstruction. Skipping.")
                continue
            fov = [r["FIELD-OF-VIEW"]["Extents"][a] for a in "XYZ"]

            attributes = {x: r[x] for x in r.keys() if x != "ReconFrames"}

            # fill in required PATATO attributes:
            pat_attributes = {}
            pat_attributes["RECONSTRUCTION_ALGORITHM"] = "iThera " + attributes["Name"]
            pat_attributes["RECONSTRUCTION_NX"] = ns[0]
            pat_attributes["RECONSTRUCTION_NY"] = ns[1]
            pat_attributes["RECONSTRUCTION_NZ"] = ns[2]
            pat_attributes["RECONSTRUCTION_FIELD_OF_VIEW_NX"] = fov[0]
            pat_attributes["RECONSTRUCTION_FIELD_OF_VIEW_NY"] = fov[1]
            pat_attributes["RECONSTRUCTION_FIELD_OF_VIEW_NZ"] = fov[2]
            pat_attributes["RECONSTRUCTION_PARAMS"] = attributes
            pat_attributes["PREPROCESSING_ALGORITHM"] = "iThera " + attributes["SingalFilterType"]
            pat_attributes["speedofsound"] = attributes["TrimSpeedOfSound"]
            pat_attributes["Notes"] = "iThera Reconstruction, imported by PATATO. Speed of sound is offset from water " \
                                      "sos at the given temperature. "

            rec = Reconstruction(recon, self._get_wavelengths(),
                                 attributes=pat_attributes,
                                 field_of_view=fov, hdf5_sub_name="iThera " + attributes["Name"])
            reconstructions.append(rec)
        rec_dict = {}
        n_rec = {}
        for r in reconstructions:
            r_name = r.attributes["RECONSTRUCTION_ALGORITHM"]
            if r_name not in rec_dict:
                n_rec[r_name] = 0
            i = n_rec[r_name]
            n_rec[r_name] += 1
            rec_dict[(r_name, str(i))] = r
        if rec_dict:
            return {"recons": rec_dict}
        else:
            return {}

    def get_speed_of_sound(self):
        return None

    def __init__(self, folder, scan_name=None):
        super().__init__()
        if scan_name is None:
            scan_name = split(folder)[-1]
        self.scan_folder = folder
        self.xml_file = join(folder, scan_name + ".msot")
        self.xml_tree = xml.dom.minidom.parse(self.xml_file)
        # Start by establishing the number of frames. This can sometimes not be exactly what is expected
        # e.g. if you opened the scanner before it had finished one sweep.
        self._ithera_get_wavelengths()
        self.nwavelengths = len(self.wavelengths)

        v24_irf = self.xml_tree.getElementsByTagName("ImpulseResponse")
        if len(v24_irf) != 0 and len(v24_irf[0].firstChild.nodeValue) == 21656:
            import base64
            self.v24_irf = np.frombuffer(base64.b64decode(v24_irf[0].firstChild.nodeValue), dtype=np.double)
        else:
            self.v24_irf = None

        self.nframes = len(self.xml_tree.getElementsByTagName("DataModelScanFrame")) // self.nwavelengths
        # Optional add here: extract the reconstructed images that are from ViewMSOT.
        # Extract attributes
        self.scan_attrs = self._get_scan_attributes()
        self.scan_elements = self._get_scan_elements()
        self.nprojections = None
        self.geometry = self.get_sensor_geometry()
        self.nsamples = self.get_n_samples()
        self.water_absorption = self.get_water_absorption()

    def get_n_samples(self):
        N_samples = int(self.xml_tree.getElementsByTagName("RECORDED-LENGTH")[0].firstChild.nodeValue)
        return N_samples

    def _get_scan_attributes(self, attrs=None):
        if attrs is None:
            attrs = [("timestamp", np.uint64), ("ultraSound-frame-offset", int)]
        output = {}
        frames = self.xml_tree.getElementsByTagName("ScanFrames")[0].getElementsByTagName("Frame")
        for attribute_tag, attribute_type in attrs:
            output[attribute_tag] = np.zeros(self.nwavelengths * self.nframes, dtype=attribute_type)
            for frame_idx, frame in enumerate(frames[:self.nwavelengths * self.nframes]):
                output[attribute_tag][frame_idx] = attribute_type(frame.getAttribute(attribute_tag))

        for attribute_tag in output:
            output[attribute_tag] = output[attribute_tag].reshape((self.nframes, self.nwavelengths))

        return output

    def _get_scan_elements(self, elements=None):
        if elements is None:
            elements = [("TEMPERATURE", float), ("POWER", float), ("DIODE-READOUT", float),
                        ("LASER-ENERGY", float), ("Z-POS", float), ("RUN", int),
                        ("REPETITION", int), ("OverallCorrectionFactor", float),
                        ]
        frames = self.xml_tree.getElementsByTagName("ScanFrames")[0].getElementsByTagName("Frame")
        output = {}
        for element_tag, element_type in elements:
            output[element_tag] = np.zeros(self.nwavelengths * self.nframes, dtype=element_type)
            for frame_idx, frame in enumerate(frames):
                if frame_idx >= self.nwavelengths * self.nframes:
                    # This means that the final frame only has some of the wavelength measurements.
                    # We will cut off this last frame.
                    break
                elements_tag = frame.getElementsByTagName(element_tag)
                if len(elements_tag) > 0:
                    value = frame.getElementsByTagName(element_tag)[0].firstChild.nodeValue
                    value = element_type(value)
                    output[element_tag][frame_idx] = value
                else:
                    if element_type == int:
                        output[element_tag][frame_idx] = -1
                    else:
                        output[element_tag][frame_idx] = np.nan

        for element_tag in output:
            output[element_tag] = output[element_tag].reshape((self.nframes, self.nwavelengths))
        return output

    def _ithera_get_wavelengths(self):
        wavelength_tree = self.xml_tree.getElementsByTagName("WAVELENGTHS")[0]
        wavelength_lookup = []
        for wl_entry in wavelength_tree.getElementsByTagName("WAVELENGTH"):
            wavelength_lookup.append(float(wl_entry.firstChild.nodeValue))
        self.wavelengths = np.array(wavelength_lookup)

    def _get_wavelengths(self):
        return self.wavelengths

    def _get_correction_factor(self):
        if not np.all(np.isnan(self.scan_elements["POWER"][()])):
            return self.scan_elements["POWER"]
        else:
            return self.scan_elements["OverallCorrectionFactor"]

    def get_impulse_response(self):
        irf_files = glob.glob(join(self.scan_folder, "*.irf"))
        if len(irf_files) > 0:
            irf_file = irf_files[0]
            return load_ithera_irf(irf_file)
        else:
            return self.v24_irf

    def _get_repetition_numbers(self):
        return self.scan_elements["REPETITION"]

    def _get_run_numbers(self):
        return self.scan_elements["RUN"]

    def get_scan_datetime(self):
        """

        Returns
        -------

        """
        import dateutil.parser
        return dateutil.parser.isoparse(
            self.xml_tree.getElementsByTagName("CreationTime")[0].firstChild.nodeValue).replace(tzinfo=None)

    def _get_scan_times(self):
        return self.scan_attrs["timestamp"]

    def _get_temperature(self):
        return self.scan_elements["TEMPERATURE"]

    def get_us_offsets(self):
        return self.scan_attrs["ultraSound-frame-offset"]

    def _get_pa_data(self):
        raw_file = glob.glob(join(self.scan_folder, "*.bin"))[0]
        raw_data = np.memmap(raw_file, mode="r", dtype=np.uint16)[
                   :self.nframes * self.nwavelengths * self.nprojections *
                    self.nsamples].reshape(self.nframes, self.nwavelengths,
                                           self.nprojections, self.nsamples)
        return raw_data, {"fs": self.get_sampling_frequency()}

    def _get_sampling_frequency(self):
        return 1e6 * float(self.xml_tree.getElementsByTagName("SAMPLING-FREQUENCY")[0].firstChild.nodeValue)

    def _get_sensor_geometry(self):
        geometry = []
        frame_desc = self.xml_tree.getElementsByTagName("FRAME-DESC")[0]
        projection_xml = frame_desc.getElementsByTagName("PROJECTION")
        for n, f in enumerate(projection_xml):
            geometry.append([])
            for i, v in enumerate(f.getElementsByTagName("VALUE")):
                geometry[-1].append(float(v.firstChild.nodeValue))
        self.nprojections = len(geometry)
        geometry = np.array(geometry)
        return geometry

    def _get_water_absorption(self):
        coeffs = []
        start = self.xml_tree.getElementsByTagName("WATER-ABSORPTION-COEFFICIENTS")[0]
        for f in start.getElementsByTagName("WATER-ABSORPTION-COEFFICIENT"):
            if f.firstChild is not None:
                coeffs.append(float(f.firstChild.nodeValue))
            else:
                coeffs.append(float(f.getAttribute("coefficient")))
        pathlength = float(self.xml_tree.getElementsByTagName("PATH-LENGTH-IN-WATER")[0].firstChild.nodeValue)
        return np.array(coeffs), pathlength

    def get_us_data(self):
        us_files = glob.glob(join(self.scan_folder, "*.us"))
        if len(us_files) > 0:
            us_nodes = self.xml_tree.getElementsByTagName("ULTRA-SOUND-FIELD-OF-VIEW")
            if len(us_nodes) > 0:
                us_node = us_nodes[0]
                N = us_node.getElementsByTagName("PixelCount")[0].getElementsByTagName("X")[0].firstChild.nodeValue
                us_pixels = int(N)
                us_extents = us_node.getElementsByTagName("Extents")
                fov = float(us_extents[0].getElementsByTagName("X")[0].firstChild.nodeValue)
            else:
                N = self.xml_tree.getElementsByTagName("ULTRA-SOUND-RESOLUTION")[0].firstChild.nodeValue
                us_pixels = int(N)
                fov = float(
                    self.xml_tree.getElementsByTagName("UltraSoundPixelSize")[0].firstChild.nodeValue) * us_pixels
            try:
                us_data = np.memmap(us_files[0], mode="r", dtype=np.float32).reshape((-1, us_pixels, us_pixels))
                us_data = np.swapaxes(us_data, -1, -2)[:, ::-1, :]
                return us_data, fov
            except ValueError:
                print("Unable to import ultrasound scans - did the scanner fail to acquire the images? SKIPPING US")
                return None
        else:
            return None, {}

    def get_scan_name(self):
        return self.xml_tree.getElementsByTagName("ScanNode")[0].getElementsByTagName("Name")[0].firstChild.nodeValue

    def _get_scanner_z_position(self):
        return self.scan_elements["Z-POS"]

    def get_scan_comment(self):
        try:
            return self.xml_tree.getElementsByTagName("Comment")[0].firstChild.nodeValue
        except AttributeError:
            return ""
        except IndexError:
            return ""



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import unittest

import numpy as np
from patato import DefaultMSOTPreProcessor, ReferenceBackprojection, PAData, iTheraMSOT
from patato.data import get_ithera_msot_time_series_example, get_msot_time_series_example
import matplotlib.pyplot as plt
import matplotlib


class TestITheraImport(unittest.TestCase):
    def test_overall_processing(self):
        # matplotlib.use('Agg')
        pa_1 = PAData(iTheraMSOT("test_data_ithera/Scan_9"))
        pa_2 = PAData.from_hdf5("test_data.hdf5")

        self.assertEqual(pa_1.get_scan_name(), pa_2.get_scan_name())
        self.assertEqual(pa_1.scan_reader.get_scan_comment(), pa_2.scan_reader.get_scan_comment())

        self.assertEqual(pa_1.get_scan_reconstructions(), {})
        pa_2.set_default_recon()

        pa_2.get_scan_reconstructions().imshow(return_scalebar_dimension=True)
        plt.close()

        us = pa_2.scan_reader.get_us_data()
        self.assertIsNone(us[0])

        time_factor = 3
        detector_factor = 2

        preproc = DefaultMSOTPreProcessor(time_factor=time_factor, detector_factor=detector_factor)
        recon = ReferenceBackprojection(field_of_view=(0.025, 0.025, 0.025), n_pixels=(333, 333, 1))

        new_t1, d1, _ = preproc.run(pa_1.get_time_series()[0:1, 0:1], pa_1[0:1, 0:1])
        rec1, _, _ = recon.run(new_t1, pa_1[0:1, 0:1], 1500, **d1)

        new_t2, d2, _ = preproc.run(pa_2.get_time_series()[0:1, 0:1], pa_2[0:1, 0:1])
        rec2, _, _ = recon.run(new_t2, pa_2[0:1, 0:1], 1500, **d2)

        self.assertTrue(np.all(pa_1.get_time_series().raw_data == pa_2.get_time_series().raw_data[:, 0]))
        self.assertTrue(np.all(new_t1.raw_data[()] == new_t2.raw_data[()]))



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import json
from typing import Union, Optional

from ..attribute_tags import PreprocessingAttributeTags, UnmixingAttributeTags, ReconAttributeTags
from ..msot_data import PAData
from ...processing.preprocessing_types import PREPROCESSING_NAMES
from ...unmixing.unmixer import SpectralUnmixer, SO2Calculator, THbCalculator
from ...recon import RECONSTRUCTION_NAMES


def read_reconstruction_preset(json_path: Union[str, dict]):
    """
        Load and parse the reconstruction preset from the specified JSON file or dictionary.

        Parameters
        ----------
        json_path : Union[str, dict]
            The path to the JSON file or a dictionary containing the reconstruction preset.

        Returns
        -------
        settings : SpatialProcessingAlgorithm
            The parsed reconstruction preset as a pipeline element.

        """
    # Load json
    if type(json_path) == str:
        with open(json_path) as json_file:
            settings = json.load(json_file)
    else:
        settings = json_path

    # Preprocessing parameters
    filter_highpass = settings.get(PreprocessingAttributeTags.HIGH_PASS_FILTER, None)
    filter_lowpass = settings.get(PreprocessingAttributeTags.LOW_PASS_FILTER, None)

    time_interpolation_factor = settings.get(PreprocessingAttributeTags.TIME_INTERPOLATION, 3)
    detector_interpolation_factor = settings.get(PreprocessingAttributeTags.DETECTOR_INTERPOLATION, 2)

    correct_for_impulse_response = settings.get(PreprocessingAttributeTags.IMPULSE_RESPONSE, True)
    do_hilbert_transform = settings.get(PreprocessingAttributeTags.HILBERT_TRANSFORM, False)
    do_envelope_detection = settings.get(PreprocessingAttributeTags.ENVELOPE_DETECTION, False)

    window_size = settings.get(PreprocessingAttributeTags.WINDOW_SIZE, 512)

    absolute_transformation = "imag"
    if do_envelope_detection:
        absolute_transformation = "abs"
    elif not do_hilbert_transform:
        absolute_transformation = "real"

    preprocessing_algorithm = PREPROCESSING_NAMES[settings[PreprocessingAttributeTags.PROCESSING_ALGORITHM]]

    # Reconstruction parameters.
    recon_params = settings.get(ReconAttributeTags.ADDITIONAL_PARAMETERS, {})  # Custom parameters

    # Field of view y and z:
    field_of_view_x = settings.get(ReconAttributeTags.X_FIELD_OF_VIEW, 0.)
    field_of_view_y = settings.get(ReconAttributeTags.Y_FIELD_OF_VIEW, 0.)
    field_of_view_z = settings.get(ReconAttributeTags.Z_FIELD_OF_VIEW, 0.)

    number_of_pixels_x = settings.get(ReconAttributeTags.X_NUMBER_OF_PIXELS, 1)
    number_of_pixels_y = settings.get(ReconAttributeTags.Y_NUMBER_OF_PIXELS, 1)
    number_of_pixels_z = settings.get(ReconAttributeTags.Z_NUMBER_OF_PIXELS, 1)

    # Number of pixels x and y
    reconstruction_algorithm = RECONSTRUCTION_NAMES[settings[ReconAttributeTags.RECONSTRUCTION_ALGORITHM]]

    # Whether to do universal backprojection.
    ubp = settings.get("UBP", False)
    step_1 = preprocessing_algorithm(time_factor=time_interpolation_factor,
                                     detector_factor=detector_interpolation_factor,
                                     irf=correct_for_impulse_response,
                                     hilbert=do_hilbert_transform,
                                     lp_filter=filter_lowpass,
                                     hp_filter=filter_highpass,
                                     filter_window_size=window_size,
                                     absolute=absolute_transformation,
                                     universal_backprojection=ubp
                                     )
    step_2 = reconstruction_algorithm((number_of_pixels_x, number_of_pixels_y, number_of_pixels_z),
                                      (field_of_view_x, field_of_view_y, field_of_view_z),
                                      **recon_params)
    step_1.add_child(step_2)
    return step_1


def read_unmixing_preset(json_path: Union[str, dict], example_data: Optional[PAData], **kwargs):
    """
    Load and parse the unmixing preset from the specified JSON file or dictionary.

    Parameters
    ----------
    json_path : Union[str, dict]
        The path to the JSON file or a dictionary containing the unmixing preset.

    example_data : PAData or None
        An example dataset used to obtain settings such as wavelengths.

    Returns
    -------
    unmixing_preset : SpatialProcessingAlgorithm
        The parsed unmixing preset as a pipeline element.

    """
    # Load json
    if type(json_path) == str:
        with open(json_path) as json_file:
            settings = json.load(json_file)
    else:
        settings = json_path
    settings.update(kwargs)

    # Load wavelengths from the preset or use the wavelengths from the example data
    wavelengths = settings.get(UnmixingAttributeTags.UNMIXING_WAVELENGTHS, None)
    wavelength_indices = settings.get(UnmixingAttributeTags.WAVELENGTH_INDICES, None)
    wavelength_range = settings.get(UnmixingAttributeTags.WAVELENGTH_RANGE, None)

    if wavelengths is None and wavelength_indices is None and wavelength_range is None:
        # Use all wavelengths from the example data
        wavelengths = example_data.get_wavelengths()[:]
    elif wavelengths is None and wavelength_indices is None:
        # Use all wavelengths within the specified range
        wavelengths = example_data.get_wavelengths()[:]
        wavelengths = wavelengths[(wavelengths >= wavelength_range[0]) & (wavelengths <= wavelength_range[1])]
    elif wavelengths is None:
        # Use the specified indices from the example data
        wavelengths = example_data.get_wavelengths()[:][wavelength_indices]

    # Get other unmixing settings from the preset
    resolution_reduction_factor = settings[UnmixingAttributeTags.RESOLUTION_REDUCE]
    spectra_names = settings[UnmixingAttributeTags.SPECTRA]

    # Initialize the spectral unmixer pipeline
    from patato.unmixing.spectra import SPECTRA_NAMES
    spectra = [SPECTRA_NAMES[x]() for x in spectra_names]

    compute_so2 = settings[UnmixingAttributeTags.COMPUTE_SO2]

    suffix = settings[UnmixingAttributeTags.SUFFIX]

    pipeline = SpectralUnmixer(spectra, wavelengths, resolution_reduction_factor, algorithm_id=suffix)

    # If haemoglobin and oxyhaemoglobin are present in the spectra, add THb and SO2 calculations to the pipeline
    if UnmixingAttributeTags.HAEMOGLOBIN in spectra_names and UnmixingAttributeTags.OXYHAEMOGLOBIN in spectra_names:
        thb = THbCalculator(suffix)
        pipeline.add_child(thb)
        if compute_so2:
            so2 = SO2Calculator(suffix)
            pipeline.add_child(so2)
    return pipeline



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import argparse
import glob
from os.path import join, split

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Slider, RangeSlider
from .. import PAData
from ..utils import sort_key
from ..core.image_structures.single_parameter_data import SingleParameterData


def init_argparse():
    parser = argparse.ArgumentParser(description="View MSOT Recons.")
    parser.add_argument('input', type=str, help="Data Folder")
    parser.add_argument('-f', '--filter', type=str, help="Choose scan", default=None)
    parser.add_argument('-r', '--recon', type=int, help="Reconstruction number", default=0)
    parser.add_argument('-fn', '--filtername', type=str, help="Choose scan name filter", default=None)
    parser.add_argument('-t', '--thb', type=bool, help="Draw on THb", default=False)
    parser.add_argument('-s', '--so2', type=bool, help="Draw on SigSo2", default=False)
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()

    DATA_FOLDER = args.input
    for file in sorted(glob.glob(join(DATA_FOLDER, "*.hdf5")), key=sort_key):
        if args.filter is not None:
            if split(file)[-1] != "Scan_" + args.filter + ".hdf5":
                continue

        data = PAData.from_hdf5(file)

        if args.filtername is not None:
            if str.lower(args.filtername) not in str.lower(data.get_scan_name()):
                continue

        print(file)
        scan_name = data.get_scan_name()

        # by default just choose the first reconstruction.
        if not args.thb and not args.so2:
            reconstructions = data.get_scan_reconstructions()
        else:
            group_name = "thb" if args.thb else "so2"
            reconstructions = data.get_scan_images(group_name, SingleParameterData)

        if reconstructions == {}:
            print(f"{file} has not been reconstructed. Skipping.")
            continue

        methods = list(reconstructions.keys())

        if len(methods) > 1:
            print("Multiple reconstructions available, using", methods[0])

        recon_data = reconstructions[methods[0]]

        extents = recon_data.extent
        recon = recon_data.raw_data

        frame_n = 0
        wl = 0

        iqr = np.nanpercentile(recon[frame_n, wl], 95) - np.nanpercentile(recon[frame_n, wl], 5)
        median = np.median(recon[frame_n, wl])
        range_interest = (median - 3 * iqr, median + 3 * iqr)

        fig = plt.figure()
        plt.subplots_adjust(bottom=0.3)
        ax, ax2 = fig.subplots(1, 2)
        p = ax.imshow(np.squeeze(recon[frame_n, wl]), extent=extents, clim=range_interest)

        ax2.hist(recon[frame_n, wl].flatten())
        vlinea = ax2.axvline(np.nanmin(recon[frame_n, wl]))
        vlineb = ax2.axvline(np.nanmax(recon[frame_n, wl]))
        ax_clims = plt.axes([0.25, 0.15, 0.5, 0.03])
        ax_slide = plt.axes([0.25, 0.11, 0.5, 0.03])
        ax_frame = plt.axes([0.25, 0.07, 0.5, 0.03])
        wavelength = Slider(ax_slide, "Wavelength", 0, recon.shape[1] - 1, valinit=0, valstep=1)
        frame = Slider(ax_frame, "Frame", 0, recon.shape[0] - 1, valinit=0, valstep=1)
        clims = RangeSlider(ax_clims, "Clim Range", range_interest[0],
                            range_interest[1], valinit=range_interest)
        ax_text = plt.axes([0.3, 0.95, 0.4, 0.04])
        ax_text.axis("off")
        ax_text.annotate(scan_name, (0.5, 0), xycoords="axes fraction",
                         annotation_clip=False, horizontalalignment="center", verticalalignment="bottom")

        def update(_):
            nonlocal frame_n, wl
            frame_n = frame.val
            wl = wavelength.val
            p.set_data(np.squeeze(recon[frame_n, wl]))
            fig.canvas.draw()

        def update_clim(val):
            cmin = val[0]
            cmax = val[1]
            vlinea.set_data(([cmin, cmin], [0, 1]))
            vlineb.set_data(([cmax, cmax], [0, 1]))
            p.set_clim([cmin, cmax])

        frame.on_changed(update)
        wavelength.on_changed(update)
        clims.on_changed(update_clim)
        plt.show()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import argparse
import glob
from os import makedirs, system
from os.path import join, split, exists


def init_argparse():
    parser = argparse.ArgumentParser(
        usage="%(prog)s [-hv] input output",
        description="Convert iThera MSOT Data into a hdf5 format. ."
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version=f"{parser.prog} version 0.1"
    )
    parser.add_argument('input', type=str, help="iThera Studies Folder")
    parser.add_argument('output', help="Empty Output Folder")
    parser.add_argument('-u', '--update', type=bool, default=False, help="Update metadata")
    parser.add_argument('-g', '--dontgetrecons', type=bool, default=False, help="Don't get Recons")
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()

    for folder in glob.glob(join(args.input, "Study_*")):
        study_name = split(folder)[-1]
        print(f"-----{study_name}-----")
        study_output = join(args.output, study_name)
        if not exists(study_output):
            makedirs(study_output)
        system(f"patato-import-ithera \"{folder}\" \"{study_output}\"" + (
            " -g True" if args.dontgetrecons else ""))



#  Copyright (c) Thomas Else 2023.
#  License: MIT
"""
``patato-set-speed-of-sound``: This module contains functions for setting the speed of sound for all datasets in a
directory.
"""

import argparse
import glob
import json
import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Slider, Button, RangeSlider

from patato.recon import get_default_recon_preset
from .. import PAData

try:
    from pyopencl.array import Array
except ImportError:
    Array = None

from .. import read_reconstruction_preset, PreProcessor,  OpenCLBackprojection
from ..recon.reconstruction_algorithm import ReconstructionAlgorithm
from ..utils import sort_key


def main():
    parser = argparse.ArgumentParser(description="Process MSOT Data.")
    parser.add_argument('input', type=str, help="Data Folder")
    parser.add_argument('c', type=float, help="Speed of Sound")
    parser.add_argument('-p', '--preset', type=str, help="Preset File",
                        default=None)
    parser.add_argument('-r', '--run', type=int, help="Run Number",
                        default=None)
    parser.add_argument('-c', '--clear', type=bool, help="Clear saved speeds",
                        default=False)
    parser.add_argument('--console', type=bool, help="Show interactive interface?",
                        default=False)
    parser.add_argument('-s', '--scan', type=str, help="Scan Number",
                        default=None)
    parser.add_argument('-b', '--startscan', type=int, help="First Scan Number",
                        default=None)
    parser.add_argument('-w', '--wavelength', type=float, help="Preview Wavelength (nm)",
                        default=None)
    parser.add_argument('-u', '--usthreshold', type=float, help="Overlay Recon on Ultrasound",
                        default=None)
    parser.add_argument('-l', '--log', type=bool, help="Log Scaling")
    parser.add_argument('-L', '--lineplots', type=bool, help="Show Line Plots")
    args = parser.parse_args()

    preset_arg = args.preset

    wavelength = args.wavelength
    if wavelength is None:
        wavelength = 800.

    data_folder = args.input
    speed_of_sound = args.c

    scan_number = args.scan
    clear = args.clear
    run_number = args.run
    if run_number is None:
        run_number = 0

    # TODO: implement ultrasound threshold overlay
    log = args.log

    files = sorted(glob.glob(os.path.join(data_folder, "**", "*.hdf5"), recursive=True), key=sort_key)

    for data_file in files:
        if args.startscan is not None:
            if int(data_file.split("_")[-1].split(".")[0]) < args.startscan:
                print("Skipping", data_file)
                continue
        if str(scan_number) not in data_file and scan_number is not None:
            continue

        print(data_file)
        pa_data = PAData.from_hdf5(data_file, "r+")[:10]
        if any(x == 0 for x in pa_data.dataset.shape):
            continue
        preset = preset_arg
        if preset_arg is None:
            preset = get_default_recon_preset(pa_data)

        with open(preset) as json_file:
            settings = json.load(json_file)

        pipeline = read_reconstruction_preset(settings)

        preprocessor: PreProcessor = pipeline
        preprocessor.time_factor = 1
        preprocessor.detector_factor = 1
        reconstructor: ReconstructionAlgorithm = pipeline.children[0]

        if pa_data.get_speed_of_sound() is not None and not clear:
            continue
        if pa_data.get_speed_of_sound() is not None:
            speed_of_sound = pa_data.get_speed_of_sound()

        if args.console:
            print(f"Setting speed of sound for {data_file}")
            pa_data.set_speed_of_sound(speed_of_sound)
            continue

        wl_index = np.argmin(np.abs(wavelength - pa_data.get_wavelengths()))
        pa_data = pa_data[:, wl_index:wl_index + 1]
        pre_processed, recon_args, _ = preprocessor.run(pa_data.get_time_series(),
                                                        pa_data)

        # print(pre_processed.raw_data.shape)
        # plt.imshow(np.squeeze(pre_processed.raw_data[0]))
        # plt.show()
        # E.g. transfer to GPU if necessary
        # pre_processed = reconstructor.pre_prepare_data(pre_processed)

        recon_info, _, _ = reconstructor.run(pre_processed[run_number:run_number + 1],
                                             pa_data[run_number:run_number + 1],
                                             speed_of_sound=speed_of_sound, **recon_args)
        recon = recon_info.raw_data

        # From here down, it's mostly Matplotlib Code.
        oa_cmap = "magma"

        extent = recon_info.extent

        fig = plt.figure(figsize=(12, 6))

        plt.suptitle(f"Tuning speed of sound for {data_file}")
        plt.subplots_adjust(bottom=0.3)
        n_plots = 3 if args.lineplots else 2
        axes = fig.subplots(1, n_plots)
        ax = axes[0]
        ax2 = axes[1]
        ax3 = None
        if n_plots == 3:
            ax3 = axes[2]

        ax_slice = plt.axes([0.25, 0.2, 0.65, 0.03])
        ax_clims = plt.axes([0.25, 0.15, 0.65, 0.03])
        ax_slide = plt.axes([0.25, 0.1, 0.65, 0.03])
        ax_button = plt.axes([0.25, 0.025, 0.4, 0.05])

        image = np.asarray(np.squeeze(recon))
        if log:
            image = np.log(image)

        image_ranges = image[np.isfinite(image) & ~np.isnan(image)]

        ax2.hist(image_ranges)
        v_line_a = ax2.axvline(np.nanmin(image))
        v_line_b = ax2.axvline(np.nanmax(image))

        p = ax.imshow(image, cmap=oa_cmap, extent=extent, origin="lower")
        xlabels = np.linspace(extent[0], extent[1], image.shape[0])
        ylabels = np.linspace(extent[2], extent[3], image.shape[1])
        if ax3 is not None:
            ax3.plot(xlabels, image[int(image.shape[0] / 2), :])
            ax3.plot(ylabels, image[:, int(image.shape[1] / 2)])

        speed = Slider(ax_slide, "Speed of Sound", 1400, 1600, valinit=speed_of_sound, valstep=1)
        image_slice = Slider(ax_slice, "Frame Number", 0, pa_data.shape[0] - 1, valinit=run_number, valstep=1)
        save_btn = Button(ax_button, "Save Speed of Sound")
        clims = RangeSlider(ax_clims, "Clim Range", np.nanmin(image_ranges), np.nanmax(image_ranges))

        def update(_):
            nonlocal pre_processed, recon_args, speed_of_sound, speed, recon, reconstructor, p, recon, log, ax3, image, pa_data
            c_new = speed.val
            recon_info, _, _ = reconstructor.run(pre_processed[image_slice.val:image_slice.val + 1],
                                                       pa_data[image_slice.val:image_slice.val + 1],
                                                       speed_of_sound=c_new,
                                                       **recon_args)
            recon = recon_info.raw_data
            image = np.squeeze(recon)
            if log:
                image = np.log(image)
            p.set_data(image)
            if ax3 is not None:
                ax3.clear()
                ax3.plot(xlabels, image[int(image.shape[0] / 2), :])
                ax3.plot(ylabels, image[:, int(image.shape[1] / 2)])
            fig.canvas.draw()

        def save(_):
            nonlocal pa_data, speed
            c_set = speed.val
            pa_data.set_speed_of_sound(c_set)
            print("Saved speed of sound")

        def update_clim(val):
            nonlocal p, v_line_a, v_line_b
            cmin = val[0]
            cmax = val[1]
            v_line_a.set_data(([cmin, cmin], [0, 1]))
            v_line_b.set_data(([cmax, cmax], [0, 1]))
            p.set_clim([cmin, cmax])

        save_btn.on_clicked(save)
        speed.on_changed(update)
        clims.on_changed(update_clim)
        image_slice.on_changed(update)

        def on_close(_):
            nonlocal pre_processed, reconstructor
            if type(pre_processed.raw_data) == Array:
                pre_processed.raw_data.data.release()
            if type(reconstructor) == OpenCLBackprojection:
                reconstructor.queue.finish()

        fig.canvas.mpl_connect('close_event', on_close)
        plt.show()


if __name__ == "__main__":
    main()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import argparse
import glob
from os.path import join

from ...io.msot_data import PAData
from ...unmixing.unmixer import DCEAnalyser
from ...utils import sort_key
from ...utils.pipeline import run_pipeline


def init_argparse():
    parser = argparse.ArgumentParser(description="Analyse Gas Challenge Data.")
    parser.add_argument('input', type=str, help="Data Folder")
    parser.add_argument('-p', '--prefix', type=str, help="Gas Challenge name prefix")
    parser.add_argument('-w', '--window', type=int, help="Smoothing window size",
                        default=10)
    parser.add_argument('-b', '--buffer', type=int, help="Buffer around changes",
                        default=5)
    parser.add_argument('-d', '--display', type=bool, help="Display steps",
                        default=True)
    parser.add_argument('-s', '--sigma', type=float, help="Smoothing window sigma",
                        default=4)
    parser.add_argument('-g', '--gas', type=lambda x: -1 if "air" == x else 1,
                        help="The starting gas for the challenge. If air then standard GC, "
                             "if o2 then it switches from o2->air", default="o2")
    parser.add_argument('--skipstart', type=int, help="Skip Runs at Start",
                        default=0)
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()

    data_folder = args.input
    prefix = args.prefix

    if prefix is None:
        prefix = ""
        raise RuntimeWarning("Warning: You should usually set a prefix for this analysis.")

    for file in sorted(glob.glob(join(data_folder, "*.hdf5")), key=sort_key):
        data = PAData.from_hdf5(file, "r+")
        scan_name = data.get_scan_name().lower()
        prefices = prefix.lower().split(",")
        if not any([x in scan_name for x in prefices]):
            continue
        data.clear_dso2()
        print(scan_name, file)

        scan_unmixed = data.get_scan_unmixed()
        if not scan_unmixed:
            print("Skipped", file, "- no unmixing.")
            continue

        analyser = DCEAnalyser(args.window, args.display, args.sigma, args.buffer)
        try:
            for recon in list(scan_unmixed.keys()):
                unmixed = scan_unmixed[recon]
                run_pipeline(analyser, unmixed, data, -1, True)
        except RuntimeError:
            print("---- SCAN NOT PROCESSED (no reference region) ---")
        data.file.close()


if __name__ == "__main__":
    main()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import unittest

import numpy as np

from patato import PAData
from patato.processing.jax_preprocessing_algorithm import PreProcessor
from patato.recon import OpenCLBackprojection, ReferenceBackprojection, SlowBackprojection


class TestBackprojection(unittest.TestCase):
    def setUp(self) -> None:
        self.pa = PAData.from_hdf5("test_data.hdf5")

        self.preproc = PreProcessor(time_factor=1, detector_factor=2)
        self.filtered_time_series, self.new_settings, _ = self.preproc.run(self.pa.get_time_series(), self.pa)

    def _test_backprojector(self, reconstructor_class):
        reconstructor = reconstructor_class([333, 334, 1], [0.025, 0.025, 1.])
        r, _, _ = reconstructor.run(self.filtered_time_series, self.pa, **self.new_settings)
        self.assertEqual(r.shape, (1, 2, 1, 334, 333))
        self.assertAlmostEqual(np.mean(r[0, 0].values), 315.63669659956736, 2)

        reconstructor = reconstructor_class([1, 333, 334], [1., 0.025, 0.025])
        r, _, _ = reconstructor.run(self.filtered_time_series, self.pa, **self.new_settings)
        self.assertEqual(r.shape, (1, 2, 334, 333, 1))

        reconstructor = reconstructor_class([334, 1, 333], [0.025, 1., 0.025])
        r, _, _ = reconstructor.run(self.filtered_time_series, self.pa, **self.new_settings)
        self.assertEqual(r.shape, (1, 2, 333, 1, 334))

    def test_reference_reconstruction(self):
        self._test_backprojector(ReferenceBackprojection)

    def test_slow_backprojection(self):
        self._test_backprojector(SlowBackprojection)

    def test_opencl_reconstruction(self):
        try:
            import pyopencl
        except ImportError:
            return  # Skip test if pyopencl is not installed

        self._test_backprojector(OpenCLBackprojection)



#  Copyright (c) Thomas Else 2023.
#  License: MIT
import unittest

import numpy as np

from patato import PAData
from patato.data.get_example_datasets import get_msot_time_series_example


class TestMSOTData(unittest.TestCase):
    def test_msot_data(self):
        pa = PAData.from_hdf5("test_data.hdf5")
        # print(pa.get_scan_name(), str(pa.get_scan_datetime()))
        pa.set_default_recon()
        self.assertTrue(pa.get_scan_name() == "201001_05_GC_SS")
        self.assertTrue(str(pa.get_scan_datetime()) == "2020-10-01 12:32:32.629145")
        self.assertEqual(pa.get_n_samples(), 2030)
        self.assertEqual(pa.get_scan_reconstructions().raw_data.shape, (1, 2, 1, 333, 333))
        self.assertEqual(pa.get_scan_unmixed().raw_data.shape, (1, 2, 1, 111, 111))
        self.assertEqual(pa.get_scan_so2().raw_data.shape, (1, 1, 1, 111, 111))
        self.assertEqual(pa.get_sampling_frequency(), 4e7)
        self.assertEqual(pa.get_scan_images("not a real image"), {})
        self.assertTrue(np.all(np.squeeze(pa.get_scan_so2_time_mean().raw_data) == np.squeeze(np.mean(pa.get_scan_so2().raw_data, axis=0))))
        self.assertTrue(np.all(np.squeeze(pa.get_scan_thb_time_mean().raw_data) == np.squeeze(np.mean(pa.get_scan_thb().raw_data, axis=0))))
        pa.close()


if __name__ == "__main__":
    t = TestMSOTData()
    t.test_msot_data()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from typing import TYPE_CHECKING

import numpy as np
from shapely.validation import make_valid

from ...io.attribute_tags import ROITags
from ...utils.mask_operations import generate_mask

if TYPE_CHECKING:
    from typing import Dict
    from ...core.image_structures.image_sequence import ImageSequence
from shapely.geometry import Polygon, MultiPolygon


class ROI:
    """Class to store regions of interests.
    """
    def get_area(self):
        return self.get_polygon().area

    def get_polygon(self) -> Polygon:
        if type(self.points) in [Polygon, MultiPolygon]:
            if not self.points.is_valid:
                return make_valid(self.points)
            else:
                return self.points
        else:
            return make_valid(Polygon(self.points))

    @property
    def attributes(self) -> "Dict":
        output = {ROITags.Z_POSITION: self.z, ROITags.RUN: self.run, ROITags.REPETITION: self.repetition,
                  ROITags.ROI_NAME: self.roi_class, ROITags.ROI_POSITION: self.position,
                  ROITags.GENERATED_ROI: self.generated}
        return output

    def __init__(self, points, z_position, run, repetition, roi_class,
                 position, generated=False, ax0_index=None):
        self.points = points
        self.z = z_position
        self.run = run
        self.repetition = repetition
        self.roi_class = roi_class
        self.position = position
        self.generated = generated
        self.ax0_index = np.array(ax0_index)

    def to_mask_slice(self, image: "ImageSequence", return_selection=False):
        mask = generate_mask(self.points, image.fov[0], image.shape_2d[-1],
                             image.fov[1], image.shape_2d[-2])
        mask = mask.reshape(image.shape[-image.n_im_dim:])
        selection = slice(None, None)
        if image.ax_0_exists():
            selection = np.where(self.ax0_index[None, :] == np.atleast_1d(image.ax_0_labels)[:, None])[0]
            ret_image = image[selection]
        else:
            ret_image = image
        if not return_selection:
            return mask, ret_image
        else:
            return mask, ret_image, selection

    def plot(self, ax=None, **kwargs):
        from ..roi_operations import REGION_COLOUR_MAP, close_loop
        import matplotlib.pyplot as plt
        if ax is None:
            ax = plt.gca()
        if type(self.points) is np.ndarray:
            plot = ax.plot(close_loop(self.points)[:, 0], close_loop(self.points)[:, 1],
                           label=self.roi_class + "_" + self.position,
                           c=REGION_COLOUR_MAP[self.roi_class], **kwargs)
        else:
            x, y = self.points.exterior.coords.xy
            plot = ax.plot(x, y,
                           label=self.roi_class + "_" + self.position,
                           c=REGION_COLOUR_MAP[self.roi_class], **kwargs)
            for interior in self.points.interiors:
                x, y = interior.coords.xy
                plot.append(ax.plot(x, y,
                               label=self.roi_class + "_" + self.position,
                               c=REGION_COLOUR_MAP[self.roi_class], **kwargs))
        return plot



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import numpy as np

try:
    import simpa as sp
except ImportError:
    sp = None

from ...core.image_structures.reconstruction_image import Reconstruction
from ..attribute_tags import ReconAttributeTags
from ..hdf.fileimporter import ReaderInterface


class SimpaImporter(ReaderInterface):
    """An importer for HDF5 files created by the SIMPA toolkit.
    """
    # Currently, aiming to just support 2D
    def _get_rois(self):
        return None

    def get_speed_of_sound(self):
        f = sp.get_data_field_from_simpa_output(self.file, sp.Tags.DATA_FIELD_SPEED_OF_SOUND)
        return np.mean(f)

    def _original_shape(self):
        return self._get_segmentation().shape

    def _get_segmentation(self):
        seg = sp.get_data_field_from_simpa_output(self.file, sp.Tags.DATA_FIELD_SEGMENTATION)
        if seg.ndim == 2:
            seg = seg[:, :, None]
        return seg

    def get_scan_comment(self):
        return ""

    def _get_sampling_frequency(self):
        if "dt_acoustic_sim" not in self.file["settings"]:
            return 1
        return 1 / self.file["settings"]["dt_acoustic_sim"]

    def _simpa_get_initial_pressure(self, wavelength):
        recon = sp.get_data_field_from_simpa_output(self.file, sp.Tags.DATA_FIELD_INITIAL_PRESSURE,
                                                    wavelength)
        if recon.ndim == 2:
            recon = recon[:, :, None]
        return recon

    def _get_datasets(self):
        wavelengths = self.wavelengths
        recon_size = self._simpa_get_initial_pressure(wavelengths[0]).shape

        recon_data = np.zeros((1, len(self.wavelengths),) + recon_size)
        
        for i, w in enumerate(wavelengths):
            recon_data[0, i] = self._simpa_get_initial_pressure(w)

        attributes = {ReconAttributeTags.X_NUMBER_OF_PIXELS: recon_data.shape[-1],
                      ReconAttributeTags.Y_NUMBER_OF_PIXELS: recon_data.shape[-2],
                      ReconAttributeTags.Z_NUMBER_OF_PIXELS: recon_data.shape[-3],
                      ReconAttributeTags.X_FIELD_OF_VIEW: self._simpa_fov[0],
                      ReconAttributeTags.Y_FIELD_OF_VIEW: self._simpa_fov[1],
                      ReconAttributeTags.Z_FIELD_OF_VIEW: self._simpa_fov[2]}
        # recon_data = recon_data.swapaxes(0, 3).copy()

        recon_class = Reconstruction(recon_data, self._get_wavelengths(),
                                     attributes=dict(attributes),
                                     field_of_view=self._simpa_fov, hdf5_sub_name="initial_pressure")

        return {"recons": {("initial_pressure", "0"): recon_class}}

    def __init__(self, filename):
        super().__init__()
        self.file = sp.load_hdf5(filename)
        self.filename = filename
        
        z0, _, x0, _, y0, _ = self.file["digital_device"].get_detection_geometry().get_field_of_view_mm()/1000
        nz, ny, nx = self._original_shape()
        dx = self.file["settings"]["voxel_spacing_mm"]/1000
        self._simpa_fov = [(x0, x0+dx*nx), (y0, y0+dx*ny), (z0, z0+dx*nz)]

        self.wavelengths = self.file["settings"]["wavelengths"]

        self.nz = 1 # for backwards compatibility

    def get_scan_datetime(self):
        return 0

    def _get_pa_data(self):
        try:
            attrs = {"fs": self.get_sampling_frequency()}
            pa = np.array([sp.get_data_field_from_simpa_output(self.file, sp.Tags.DATA_FIELD_TIME_SERIES_DATA, w) for w in
                           self.wavelengths])

            return (pa[None],
                    attrs)
        except KeyError:
            return (np.full((1, len(self.wavelengths), 256, 2), np.nan), {})

    def get_scan_name(self):
        return self.filename

    def _get_temperature(self):
        return np.zeros((self.nz, len(self.wavelengths)))

    def _get_correction_factor(self):
        return np.ones((self.nz, len(self.wavelengths)))

    def _get_scanner_z_position(self):
        return np.arange(self.nz)[:, None] * np.ones((len(self.wavelengths),))[None, :] * self.file["settings"][
            "voxel_spacing_mm"]

    def _get_run_numbers(self):
        return np.zeros((self.nz, len(self.wavelengths)))

    def _get_repetition_numbers(self):
        return np.zeros((self.nz, len(self.wavelengths)))

    def _get_scan_times(self):
        return np.arange(self.nz * len(self.wavelengths)).reshape((self.nz, len(self.wavelengths)))

    def _get_sensor_geometry(self):
        g = self.file["digital_device"].get_detection_geometry().get_detector_element_positions_accounting_for_field_of_view()
        return g / 1000

    def get_us_data(self):
        return None

    def get_impulse_response(self):
        irf = np.zeros((self.get_pa_data()[0].shape[-1],))
        irf[irf.shape[0] // 2] = 1
        return irf

    def _get_wavelengths(self):
        return self.wavelengths

    def get_us_offsets(self):
        return None

    def _get_water_absorption(self):
        return np.zeros((len(self.wavelengths),)), 0



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import matplotlib.animation as animation
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap

from ..io.attribute_tags import HDF5Tags


def animate_sequence(images, cmap, clim, **kwargs):
    fig, ax = plt.subplots()

    ims = []
    for i in range(images.shape[0]):
        im = ax.imshow(images[i], animated=True, cmap=cmap, clim=clim)
        if i == 0:
            ax.imshow(images[i], cmap=cmap, clim=clim)
        ims.append([im])

    default_kwargs = {"interval": 50, "blit": True, "repeat_delay": 1000}
    for d, k in default_kwargs.items():
        if d not in kwargs:
            kwargs[d] = k

    ani = animation.ArtistAnimation(fig, ims, **kwargs)
    return ani


type_cmaps = {HDF5Tags.SO2: "RdBu_r",
              HDF5Tags.THB: "magma",
              HDF5Tags.DELTA_SO2: "viridis",
              HDF5Tags.BASELINE_SO2: "RdBu_r",
              HDF5Tags.BASELINE_SO2_STANDARD_DEVIATION: "plasma",
              HDF5Tags.RECONSTRUCTION: "bone",
              HDF5Tags.DELTA_ICG: "cividis",
              "Responding Pixels": ListedColormap([(0, 0, 0, 0), "orange"]),
              HDF5Tags.RAW_DATA: "bone",
              None: "viridis",
              HDF5Tags.UNMIXED: "viridis"}



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import argparse

import h5py


def init_argparse():
    parser = argparse.ArgumentParser(description="Rename MSOT Scans.")
    parser.add_argument('input_file', type=str, help="Input file name")
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()
    data = h5py.File(args.input_file, "r+")
    name = data.attrs.get("name", None)
    if "name" not in data.attrs:
        name = data["raw_data"].attrs.get("name")
    original_name = data.attrs.get("original_name", None)
    if original_name is None:
        original_name = name
        data.attrs["original_name"] = name
    print(f"Current name: {name}. Original name: {original_name}.")
    while not (question := input("What would you like to change the name to? ")):
        print("Please enter a valid name.")

    data.attrs["name"] = question
    data["raw_data"].attrs["name"] = question



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
PATATO command line interfaces
==============================

A series of convenience scripts for running the patato package in the command line.
"""



"""
pa_time_data. Defines the time-domain version of PARawData class.
"""

#  Copyright (c) Thomas Else 2023.
#  License: MIT


import numpy as np
import xarray

from ..image_structures.pa_raw_data import PARawData


try:
    from pyopencl.array import Array, to_device
except ImportError:
    Array = np.ndarray
    to_device = None


class PATimeSeries(PARawData):
    """
    PATimeSeries is the data structure for time-domain raw PA data.
    """

    def __add__(self, other):
        new_data = xarray.concat([self.da, other.da], dim=other.da.dims[0])
        output = PATimeSeries(new_data, new_data.dims, new_data.coords, new_data.attrs,
                               self.hdf5_sub_name, self.algorithm_id)
        output.__class__ = self.__class__
        return output

    def get_hdf5_group_name(self):
        return "raw_data"

    def to_opencl(self, queue) -> "PATimeSeries":
        cls = self.copy()
        cls.da = cls.da.copy()
        new_data = to_device(queue, self.raw_data.astype(np.single))

        # Fudge to make pyopencl ducktyped with numpy array
        new_data.__array_function__ = None
        new_data.__array_ufunc__ = None

        cls.raw_data = new_data
        return cls

    def two_dims(self):
        return "detectors", "timeseries"



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import unittest

import numpy as np

from patato import Reconstruction, SO2Calculator, SpectralUnmixer, THbCalculator, PAData
from patato.data.get_example_datasets import get_msot_time_series_example


class TestUnmixing(unittest.TestCase):
    def test_numpy_unmix(self):
        image = np.zeros((1, 2, 333, 333, 1))
        image[:, 0] += 1112  # Manually chosen values to match the absorption of 650 nm and 800 nm for 78% oxygenated Hb
        image[:, 1] += 804
        wavelengths = np.array([650, 800])

        r = Reconstruction(image, wavelengths,
                           field_of_view=(1, 1, 1))  # field of view is the width of the image along x, y, z

        r.attributes["RECONSTRUCTION_FIELD_OF_VIEW_X"] = 1
        r.attributes["RECONSTRUCTION_FIELD_OF_VIEW_Y"] = 1
        r.attributes["RECONSTRUCTION_FIELD_OF_VIEW_Z"] = 1

        um = SpectralUnmixer(["Hb", "HbO2"], r.wavelengths)
        so = SO2Calculator()

        um, _, _ = um.run(r, None)
        so2, _, _ = so.run(um, None)

        self.assertAlmostEqual(np.mean(so2.raw_data), 0.7799958428309177)

    def test_unmix(self):
        pa = PAData.from_hdf5("test_data.hdf5")

        pa.set_default_recon()
        r = pa.get_scan_reconstructions()

        unmixer = SpectralUnmixer(["Hb", "HbO2"], pa.get_wavelengths())

        u, _, _ = unmixer.run(r, pa)

        so2_calc = SO2Calculator()
        s, _, _ = so2_calc.run(u, pa)

        thb_calc = THbCalculator()
        t, _, _ = thb_calc.run(u, pa)
        self.assertTrue(np.all(pa.get_scan_so2_time_mean().raw_data == np.mean(pa.get_scan_so2().raw_data,
                                                                               axis=(0, 1))))
        self.assertTrue(np.all(pa.get_scan_so2_time_standard_deviation().raw_data == np.std(pa.get_scan_so2().raw_data,
                                                                                            axis=(0, 1))))

        # Get these to do sanity checks - better to implement proper tests in future.
        self.assertIsNone(pa.get_segmentation())
        self.assertIsNotNone(pa.get_recon_types())
        self.assertIsNotNone(pa.get_recon_types())
        self.assertIsNotNone(pa.get_scan_dso2())
        self.assertIsNotNone(pa.get_scan_baseline_so2())
        self.assertIsNotNone(pa.get_scan_baseline_standard_deviation_so2())
        self.assertIsNotNone(pa.get_scan_dicg())
        self.assertIsNotNone(pa.get_scan_baseline_icg())
        self.assertIsNotNone(pa.get_scan_baseline_standard_deviation_icg())
        self.assertIsNone(pa.get_responding_pixels())
        self.assertIsNotNone(pa.get_rois())
        self.assertIsNotNone(pa.summary_measurements())

        self.assertEqual(u.shape, (1, 2, 1, 333, 333))
        self.assertEqual(s.shape, (1, 1, 1, 333, 333))
        self.assertEqual(t.shape, (1, 1, 1, 333, 333))


if __name__ == "__main__":
    unittest.main()
    print("Tests complete.")



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import logging
import os
from typing import Sequence

import numpy as np

try:
    import pyopencl as cl
    import pyopencl.array
except ImportError:
    cl = None
    pyopencl = None

from .reconstruction_algorithm import ReconstructionAlgorithm
from ..core.image_structures.pa_time_data import PATimeSeries


def validate_opencl_input(signal: np.ndarray, fs, detectors, n_pixels, field_of_view, speed_of_sound):
    # Make sure that the input to the OpenCL code is valid.
    if type(signal) is not np.ndarray:
        signal = np.asarray(signal)
    if type(detectors) is not np.ndarray:
        detectors = np.asarray(detectors)

    if not signal.flags["C_CONTIGUOUS"]:
        signal = signal.copy()
        logging.warning("Warning: order of signal is not C_CONTIGUOUS, this might hinder performance.")
    if not (type(fs) == float or type(fs) in [np.float64, np.float32]):
        fs = float(fs)
    if not detectors.flags["C_CONTIGUOUS"]:
        detectors = detectors.copy()
    assert type(n_pixels[0]) in [int, np.int32]
    field_of_view = tuple(float(x) for x in field_of_view)
    if type(speed_of_sound) not in [float, np.float32, np.float64]:
        speed_of_sound = float(speed_of_sound)
    return signal, fs, detectors, n_pixels, field_of_view, speed_of_sound


class OpenCLBackprojection(ReconstructionAlgorithm):
    """
    An OpenCL implemntation of the photoacoustic backprojection.

    Note: this class will be deprecated in future. It is maintained for backwards compatibility with non-nvidia
    GPUs.
    """
    def __init__(self,
                 n_pixels: Sequence[int],
                 field_of_view: Sequence[float],
                 **kwargs):
        self.ctx = kwargs.get("ctx", None)
        self.queue = kwargs.get("queue", None)
        if self.ctx is None:
            self.ctx = pyopencl.create_some_context(interactive=False)
            self.queue = pyopencl.CommandQueue(self.ctx)

        # Connect to the OpenCL code:
        directory = os.path.dirname(os.path.abspath(__file__))

        with open(os.path.join(directory, "opencl_code/dascl.cl")) as file:
            code = file.read()

        program = cl.Program(self.ctx, code).build()
        cldas = program.cldas
        cldas.set_scalar_arg_dtypes([None, None, None, np.float32, np.intc, np.intc, np.intc,
                                     np.float32, np.float32, np.float32, np.float32, np.float32,
                                     np.float32, np.intc, np.intc, None, np.intc, np.intc])
        super().__init__(n_pixels, field_of_view, ctx=self.ctx, queue=self.queue, cldas=cldas)

    def pre_prepare_data(self, x: PATimeSeries):
        # really just for speed of sound setting
        return x.to_opencl(self.queue)

    def reconstruct(self, signal, fs, detectors, n_pixels: Sequence[int],
                    field_of_view: Sequence[float],
                    speed_of_sound=None,
                    queue=None,
                    cldas=None,
                    **kwargs) -> np.ndarray:
        signal, fs, detectors, n_pixels, field_of_view, speed_of_sound = validate_opencl_input(signal, fs, detectors,
                                                                                               n_pixels, field_of_view,
                                                                                               speed_of_sound)
        if type(signal) is pyopencl.array.Array and queue is None:
            queue = signal.queue
        if queue is None:
            ctx = cl.create_some_context(interactive=False)
            queue = cl.CommandQueue(ctx)
        if cldas is None:
            raise ValueError()

        if type(signal) is not pyopencl.array.Array:
            signal = pyopencl.array.to_device(queue, signal.astype(np.float32))

        logging.info(f"{detectors.flags}")
        if type(detectors) is not pyopencl.array.Array:
            detectors = pyopencl.array.to_device(queue, detectors.copy().astype(np.float32))

        logging.info(f"{signal.shape}, {speed_of_sound}, {fs}, {detectors.shape}")

        original_shape = signal.shape[:-2]
        frames = int(np.product(original_shape))
        signal = signal.reshape((frames,) + signal.shape[-2:])
        dl = speed_of_sound / fs

        logging.info(f"{fs}, {speed_of_sound}.")

        nx, ny, nz = n_pixels

        if nx == 1:
            dx = 1
        else:
            dx = field_of_view[0] / (n_pixels[0] - 1)

        if ny == 1:
            dy = 1
        else:
            dy = field_of_view[1] / (n_pixels[1] - 1)

        if nz == 1:
            dz = 1
        else:
            dz = field_of_view[2] / (n_pixels[2] - 1)

        x_0 = - field_of_view[0] / 2 if nx != 1 else 0
        y_0 = - field_of_view[1] / 2 if ny != 1 else 0
        if nz == 1:
            z_0 = 0
        else:
            z_0 = - field_of_view[2] / 2

        n_detectors = detectors.shape[0]
        n_samples = signal.shape[-1]

        max_size = min(queue.device.get_info(cl.device_info.MAX_WORK_GROUP_SIZE), n_detectors)
        local_memory = cl.LocalMemory(max_size * 4)
        frame_offset = int(signal.offset // (n_detectors * n_samples * 4))
        output = pyopencl.array.zeros(queue, (frames, n_detectors // max_size) + tuple(n_pixels), dtype=signal.dtype)

        logging.info(f"OpenCL debugging: {queue}, {nx * ny * nz}, {n_detectors}, {max_size}, {output.dtype}, "
                     f"{signal.dtype}, {detectors.dtype}, {dl}, {nx}, {ny}, {nz}, {x_0}, "
                     f"{y_0}, {z_0}, {dx}, {dy}, {dz}, {n_detectors}, {n_samples}, "
                     f"{local_memory}, {frames}, {frame_offset}, {signal.shape}, {original_shape}.")
        for f in range(frames):
            logging.info(f"Step {f}.")
            try:
                cldas(queue, (nx * ny * nz, n_detectors), (1, max_size),
                      output.data, signal.base_data, detectors.data, dl,
                      nx, ny, nz, x_0, y_0, z_0, dx, dy, dz, n_detectors,
                      n_samples, local_memory,
                      f, frame_offset)
            except pyopencl.LogicError:
                max_size = 1
                local_memory = cl.LocalMemory(max_size * 4)
                output = pyopencl.array.zeros(queue, (frames, n_detectors // max_size) + tuple(n_pixels),
                                              dtype=signal.dtype)
                cldas(queue, (nx * ny * nz, n_detectors), (1, max_size),
                      output.data, signal.base_data, detectors.data, dl,
                      nx, ny, nz, x_0, y_0, z_0, dx, dy, dz, n_detectors,
                      n_samples, local_memory,
                      f, frame_offset)
            queue.finish()

        numpy_output = np.sum(output.get(), axis=1)
        try:
            output.base_data.release()
        except AttributeError:
            logging.debug("OpenCL can't release base data, continuing anyway...")
        queue.flush()
        return numpy_output.reshape(original_shape + tuple(n_pixels)[::-1])

    @staticmethod
    def get_algorithm_name() -> str:
        return "OpenCL Backprojection"



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import argparse
import glob
from os.path import split, join

from .. import PAData
from ..utils import sort_key
from ..utils.roi_operations import add_rim_core_data


def init_argparse():
    parser = argparse.ArgumentParser(description="Split regions of interest.")
    parser.add_argument('input', type=str, help="Data Folder")
    parser.add_argument('name', type=str, help="ROI Name")
    parser.add_argument('number', type=str, help="ROI Number")
    parser.add_argument('distance', type=float, help="Core Distance (in metres)")
    parser.add_argument('-r', '--radius', type=bool, help="Distance is desired radius", default=False)
    parser.add_argument('-f', '--filter', type=str, help="Choose scan", default=None)
    parser.add_argument('-fn', '--filtername', type=str, help="Choose scan name filter", default=None)
    parser.add_argument('-c', '--clear', type=str, help="Clear all generated ROIs", default=False)
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()

    data_folder = args.input

    for file in sorted(glob.glob(join(data_folder, "**", "*.hdf5"),recursive=True), key=sort_key):
        if args.filter is not None:
            if split(file)[-1] != "Scan_" + args.filter + ".hdf5":
                continue
        data = PAData.from_hdf5(file, "r+")

        if args.filtername is not None:
            if args.filtername.upper() not in data.get_scan_name().upper():
                continue

        if args.clear and "rois" in data.file:
            for name in data.file["rois"]:
                for num in data.file["rois"][name]:
                    if data.file["rois"][name][num].attrs.get("generated", False):
                        del data.file["rois"][name][num]

        if (args.name, args.number) in data.get_rois():
            print(file)
            new_roi, _ = add_rim_core_data(data, (args.name, args.number), args.distance, radius=args.radius)
            data.add_roi(new_roi, generated=True)
            data.file.close()


if __name__ == "__main__":
    main()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
Unmixing module
======================

This module provides functions for commonly-used spectral analysis procedures in photoacoustic imaging.
Additionally, the module provides commonly-used spectra for chromophores in biological tissue.
"""


def get_default_unmixing_preset():
    """

    Returns
    -------

    """
    import os
    root_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "../unmixing/unmix_presets")
    return os.path.join(root_folder, "haemoglobin.json")



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import pybind11
import setuptools
from setuptools import Extension

modelbased_c = Extension("patato.recon.model_based.generate_model",
                         sources=["patato/recon/model_based/generate_model.cpp"],
                         language="c++", extra_compile_args=["-std=c++11"],
                         include_dirs=[pybind11.get_include()], )
modelbased_c2 = Extension("patato.recon.model_based.generate_model_refraction",
                          sources=["patato/recon/model_based/generate_model_refraction.cpp"],
                          language="c++", extra_compile_args=["-std=c++11"],
                          include_dirs=[pybind11.get_include()], )

setuptools.setup(
    ext_modules=[modelbased_c2, modelbased_c]
)



#  Copyright (c) Thomas Else 2023.
#  Copyright (c) Janek Grohl 2023.
#  License: MIT

from os.path import split
import pacfish as pf
import numpy as np
from ..hdf.fileimporter import ReaderInterface


class IPASCInterface(ReaderInterface):
    """
    An interface for iThera MSOT datasets.
    """
    def _get_rois(self):
        pass

    def _get_segmentation(self):
        return None

    def _get_datasets(self):
        return None

    def get_speed_of_sound(self):
        return None

    def __init__(self, file_path):
        super().__init__()
        self.scan_name = split(file_path)[-1]
        self.pa_data = pf.load_data(file_path)
        self.nwavelengths = len(self.pa_data.get_acquisition_wavelengths())
        self.nframes = np.shape(self.pa_data.binary_time_series_data)[3]
        # Optional add here: extract the reconstructed images that are from ViewMSOT.
        # Extract attributes
        self.geometry = self.pa_data.get_detector_position()
        self.nsamples = self.nwavelengths * self.nframes

    def get_n_samples(self):
        return self.nsamples

    def _get_wavelengths(self):
        return self.pa_data.get_acquisition_wavelengths()

    def _get_correction_factor(self):
        return self.pa_data.get_overall_gain()

    def get_impulse_response(self):
        return self.pa_data.get_frequency_response()

    def _get_repetition_numbers(self):
        return self.nsamples

    def _get_run_numbers(self):
        return np.arange(1, self.nframes+1)

    def get_scan_datetime(self):
        return None

    def _get_scan_times(self):
        return self.pa_data.get_measurement_time_stamps()

    def _get_temperature(self):
        return self.pa_data.get_temperature()

    def get_us_offsets(self):
        return None

    def _get_pa_data(self):
        raw_data = self.pa_data.binary_time_series_data # [detectors, samples, wavelengths, frames]
        raw_data = np.swapaxes(raw_data, 1, 3)  # [detectors, frames, wavelengths, samples]
        raw_data = np.swapaxes(raw_data, 0, 2)  # [wavelengths, frames, detectors, samples]
        raw_data = np.swapaxes(raw_data, 0, 1)  # [frames, wavelengths, detectors, samples]
        return raw_data, {"fs": self.pa_data.get_sampling_rate()}

    def _get_sampling_frequency(self):
        return self.pa_data.get_sampling_rate()

    def _get_sensor_geometry(self):
        return self.pa_data.get_detector_position()

    def _get_water_absorption(self):
        return None

    def get_us_data(self):
        return None

    def get_scan_name(self):
        return self.scan_name

    def _get_scanner_z_position(self):
        return self.pa_data.get_measurement_spatial_poses()

    def get_scan_comment(self):
        return ""



"""
This defines the data structure for reconstructed images.
"""

#  Copyright (c) Thomas Else 2023.
#  License: MIT

from __future__ import annotations

import numpy as np

from ...core.image_structures.image_sequence import ImageSequence
from ...io.attribute_tags import HDF5Tags


class Reconstruction(ImageSequence):
    """
    Data structure for reconstructed images.
    """
    save_output = True

    @staticmethod
    def is_single_instance():
        return False

    @staticmethod
    def get_ax1_label_meaning():
        return HDF5Tags.WAVELENGTH

    def get_hdf5_group_name(self) -> str:
        return HDF5Tags.RECONSTRUCTION

    @property
    def wavelengths(self):
        return np.array(self.ax_1_labels)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from collections import defaultdict
from typing import TYPE_CHECKING, Tuple

import numpy as np

if TYPE_CHECKING:
    from ..io.msot_data import PAData
    from .rois.roi_type import ROI

from seaborn.palettes import color_palette


def split_roi_left_right(data: "PAData", base_roi="", split_template="unnamed"):
    from .rois.roi_type import ROI
    # Initial implementation - only supports one "dividing" roi.
    data_rois = data.get_rois()

    template_data = None
    names = None
    if (split_template + "_left", "0") in data_rois:
        template_data = data_rois[(split_template + "_left", "0")]
        names = ["left", "right"]
    elif (split_template + "_right", "0") in data_rois:
        template_data = data_rois[(split_template + "_right", "0")]
        names = ["right", "left"]
    else:
        return []

    template_polygon = template_data.get_polygon()
    output = []
    for (name, number), roi in data_rois.items():
        if roi.position != "" or roi.generated:
            continue
        if base_roi in name:
            # Process
            base_polygon = roi.get_polygon()
            base_region_a = base_polygon & template_polygon
            base_region_b = base_polygon - base_region_a
            # Convert to numpy array
            try:
                x, y = base_region_a.exterior.coords.xy
                region_a = np.array([x[:-1], y[:-1]]).T
                x, y = base_region_b.exterior.coords.xy
                region_b = np.array([x[:-1], y[:-1]]).T
            except AttributeError:
                continue

            # Get the metadata.
            z_position = roi.z
            run = roi.run
            repetition = roi.repetition
            roi_class = roi.roi_class

            roi_a = ROI(region_a, z_position, run, repetition, roi_class,
                        names[0])
            roi_b = ROI(region_b, z_position, run, repetition, roi_class,
                        names[1])
            output.append(roi_a)
            output.append(roi_b)
    return output


ROI_NAMES = ["brain", "body", "reference", "aorta", "tumour", "background", "artery",
             "vein", "muscle", "phantom", "unnamed", "kidney", "spleen", "spine"]

REGION_COLOURS = color_palette("husl", len(ROI_NAMES))

REGION_COLOUR_MAP = defaultdict(lambda: REGION_COLOURS[-1])
for x, y in zip(ROI_NAMES, REGION_COLOURS):
    REGION_COLOUR_MAP[x] = y


def close_loop(x):
    return np.concatenate([x, x[0:1, :]])


def get_rim_core_rois(roi_data: "ROI", distance: float, radius=False):
    from ..io.msot_data import ROI
    roi = roi_data.get_polygon()
    if radius:
        effective_radius = np.sqrt(roi.area / np.pi)
        # distance is the desired radius
        distance = -distance + effective_radius

    roi_buffer = roi.buffer(-distance)

    core_roi = ROI(roi_buffer, roi_data.z, roi_data.run, roi_data.repetition, roi_data.roi_class + ".core",
                   roi_data.position, generated=True, ax0_index=roi_data.ax0_index)
    rim_roi = ROI(roi - roi_buffer, roi_data.z, roi_data.run, roi_data.repetition, roi_data.roi_class + ".rim",
                  roi_data.position, generated=True, ax0_index=roi_data.ax0_index)
    return core_roi, rim_roi


def add_rim_core_data(data: "PAData", base_roi: Tuple[str, str], distance: float, radius=False):
    roi_data: "ROI" = data.get_rois()[base_roi]
    return get_rim_core_rois(roi_data, distance, radius)



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import os
import shutil
import zipfile
from tempfile import mkdtemp
from tqdm.auto import tqdm

import requests
from patato import iTheraMSOT

from .. import PAData


def get_patato_data_folder():
    """Get the folder where PATATO example data is stored.

    Returns
    -------
    folder : str
        The folder where patato data is stored.
    """
    folder = os.environ.get("PAT_DATA_FOLDER") or "~/patato_example_data"

    if folder == "TEMP":
        os.environ["PAT_DATA_FOLDER"] = mkdtemp()
        folder = os.environ["PAT_DATA_FOLDER"]
    return os.path.expanduser(folder)


def download_file(file_from, file_to):
    raise NotImplementedError("No longer able to automatically download data."
                              "Please download manually from: https://doi.org/10.17863/CAM.93181")


def get_msot_time_series_example(image_type="so2"):
    """Get a time series of MSOT images.

    Returns
    -------
    dataset : PAData
        The MSOT dataset.
    """

    data_sources = {"so2": "https://www.repository.cam.ac.uk/bitstream/handle/1810/345836/invivo_oe.hdf5",
                    "icg": "https://www.repository.cam.ac.uk/bitstream/handle/1810/345836/invivo_dce.hdf5"}

    data_path = os.path.join(get_patato_data_folder(), f'{image_type}-timeseries-data.hdf5')
    folder = os.path.split(data_path)[0]
    os.makedirs(folder, exist_ok=True)
    if not os.path.exists(data_path):
        # Download the data
        download_file(data_sources[image_type], data_path)
    return PAData.from_hdf5(data_path)


def get_ithera_msot_time_series_example(image_type="so2"):
    """Get a time series of MSOT images in the iThera format.

    Returns
    -------
    dataset : PAData
        The MSOT dataset.
    """
    import glob
    print(glob.glob(get_patato_data_folder() + "/*"))
    data_sources = {"so2": "https://www.repository.cam.ac.uk/bitstream/handle/1810/345836/ithera_invivo_oe.zip",
                    "icg": "https://www.repository.cam.ac.uk/bitstream/handle/1810/345836/ithera_invivo_dce.zip"}

    data_path = os.path.join(get_patato_data_folder(), f'{image_type}-ithera_data')
    filenames = {"so2": "Scan_9", "icg": "Scan_10"}
    folder = os.path.split(data_path)[0]
    if folder and not os.path.exists(folder):
        os.makedirs(folder, exist_ok=True)
    if not os.path.exists(data_path):
        # Download the data
        zip_file = os.path.join(get_patato_data_folder(), "patato_temp.zip")
        download_file(data_sources[image_type], zip_file)
        with zipfile.ZipFile(zip_file, 'r') as zip_ref:
            zip_ref.extractall(data_path)
        print(f"Extracted data to {data_path}")
    return PAData(iTheraMSOT(os.path.join(data_path, filenames[image_type])))


def get_msot_phantom_example(image_type="clinical"):
    """Get a time series of MSOT images.

    Returns
    -------
    dataset : PAData
        The MSOT dataset.
    """
    data_sources = {"clinical": "https://www.repository.cam.ac.uk/bitstream/handle/1810/345836/clinical_phantom.hdf5",
                    "preclinical": "https://www.repository.cam.ac.uk/bitstream/handle/1810/345836/preclinical_phantom.hdf5"}

    data_path = os.path.join(get_patato_data_folder(), f'{image_type}-msot-data.hdf5')
    folder = os.path.split(data_path)[0]
    if not os.path.exists(folder):
        os.makedirs(folder, exist_ok=True)
    if not os.path.exists(data_path):
        # Download the data
        download_file(data_sources[image_type], data_path)
    return PAData.from_hdf5(data_path)



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import argparse
import glob
from os.path import join

import h5py
from ..utils import sort_key


def init_argparse():
    parser = argparse.ArgumentParser(description="Look at the status of a data folder.")
    parser.add_argument('input', type=str, help="Data Folder")
    parser.add_argument(
        "-v", "--version", action="version",
        version=f"{parser.prog} version 0.1"
    )
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()

    for filename in sorted(glob.glob(join(args.input, "*.hdf5")), key=sort_key):
        print(filename)
        data_file = h5py.File(filename, "r")

        # 1. Check raw data and print name

        if "raw_data" in data_file:
            print("Scan name", data_file["raw_data"].attrs["name"])
        print("Raw data present" if "raw_data" in data_file else "Raw data not present")
        if "raw_data" in data_file:
            print("Speed of sound set" if "speedofsound" in data_file["raw_data"].attrs else "Speed of sound not set")

        # 2. Check reconstructions

        if "recons" in data_file:
            for method in data_file["recons"]:
                print(method + ": " + str(len(data_file["recons"][method])))
        else:
            print("No reconstructions present")

        # 3. Unmixing?

        if "unmixed" in data_file:
            for method in data_file["unmixed"]:
                print("Unmixed results:" + method + ": " + str(len(data_file["unmixed"][method])))
        else:
            print("No unmixed results present")

        # 4. Rois?

        if "rois" in data_file:
            for region in data_file["rois"]:
                print(region + ":", len(data_file["rois"][region]))
        else:
            print("No rois available")
        print("\n")



"""
This defines the data structure for single parameter datasets like sO2 and THb.
"""

#  Copyright (c) Thomas Else 2023.
#  License: MIT

from ..image_structures.image_sequence import ImageSequence


class SingleParameterData(ImageSequence):
    """
    SingleParameterData is the datastructure for images like sO2 and THb (one value per frame per pixel).
    """
    save_output = True
    @staticmethod
    def is_single_instance():
        return False

    @staticmethod
    def get_ax1_label_meaning():
        return "parameter"

    def get_hdf5_group_name(self):
        return self.parameters[0].item()

    @property
    def parameters(self):
        return self.ax_1_labels

    def __init__(self, raw_data, ax_1_labels=None, algorithm_id="", field_of_view=None, attributes=None,
                 hdf5_sub_name=None):
        if len(ax_1_labels) != 1 or raw_data.shape[1] != 1:
            raise ValueError("Single parameter data requires raw_data.shape[1] to be 1.")
        super().__init__(raw_data, ax_1_labels, algorithm_id, field_of_view, attributes, hdf5_sub_name)



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import unittest
from os.path import join, split

from patato import SpectralUnmixer, PreProcessor
from patato.io.json.json_reading import read_reconstruction_preset, read_unmixing_preset
from patato.processing.preprocessing_algorithm import NumpyPreProcessor
from patato.processing.gpu_preprocessing_algorithm import GPUMSOTPreProcessor


class TestJSONLoading(unittest.TestCase):
    def test_load_recon_preset(self):
        folder, _ = split(__file__)
        pipeline = read_reconstruction_preset(join(folder,
                                                   "../patato/recon/recon_presets/backproject_standard_xz.json"))

        self.assertIn(type(pipeline), [PreProcessor, NumpyPreProcessor, GPUMSOTPreProcessor])

    def test_load_unmixed_preset(self):
        folder, _ = split(__file__)
        pipeline = read_unmixing_preset(join(folder, "../patato/unmixing/unmix_presets/haemoglobin.json"), None,
                                        WAVELENGTHS=[700, 900])

        self.assertEqual(type(pipeline), SpectralUnmixer)


if __name__ == '__main__':
    unittest.main()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import argparse
import glob
import re
from os.path import join

import h5py
import numpy as np
from ..utils import sort_key


def init_argparse():
    map = {"TRUE": True, "FALSE": False, "T": True, "F": False, "YES": True, "NO": False, "Y": True, "N": False}
    def map_fn(x):
        return map[x.upper()]
    parser = argparse.ArgumentParser(description="Copy ROIs between scans made of same thing at same time.")
    parser.add_argument('input', type=str, help="Data Folder")
    parser.add_argument('-r', '--regex', default=None, type=str, help="Regex for parsing name")
    parser.add_argument('-c', '--copyclose', default=False, type=map_fn, help="Automatically copy over roi to closest "
                                                                              "slice if not exact matches (tolerance "
                                                                              "1mm)")
    parser.add_argument('-d', '--deleteold', default=True, type=map_fn, help="Delete old copies.")
    parser.add_argument('-j', '--justdeletecopies', default=False, type=map_fn,
                        help="Just delete copies of rois.")
    parser.add_argument('-f', '--copyfrom', default=None,
                        help="Copy only from scan type (e.g. OE).")
    parser.add_argument('-t', '--copyto', default=None,
                        help="Copy only to scan type (e.g. DCE).")
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()

    DATA_FOLDER = args.input

    regex = args.regex

    if args.regex is None:
        regex = r"^(?P<mouse>[^_\- ]+)[_\- ](?P<ear>[^_\- ]+)[_\- ]((?P<scan_number>[0-9]+)[_\- ])?(?P<scan_type>.+)$"
    elif args.regex == "tom":
        regex = r"^(?P<date>[0-9]+)\-(?P<mouse>.+)\-(?P<scan_type>.+)"
    elif args.regex == "marilena":
        regex = r"VMstudy_MEO#(?P<number>[0-9]+)_((Day(?P<day>[0-9]+))|MSOT)_?(?P<scan_type>GS|SS|MS|GC|again)?"

    regex = re.compile(regex)

    scan_groups = {}

    for file in sorted(glob.glob(join(DATA_FOLDER, "**", "*.hdf5"), recursive=True), key=sort_key):
        data = h5py.File(file, "r")
        name_regex = regex.fullmatch(data["raw_data"].attrs["name"])
        print(file, data["raw_data"].attrs["name"], name_regex)
        if name_regex is None:
            continue
        name_regex = name_regex.groupdict()
        scan_group = []
        for k in name_regex:
            if k != "scan_type":
                if name_regex[k] is not None:
                    scan_group.append(name_regex[k])
                else:
                    scan_group.append("1")
        scan_group = "-".join(scan_group)
        if scan_group not in scan_groups:
            scan_groups[scan_group] = [file]
        else:
            scan_groups[scan_group].append(file)
        data.close()

    for group in scan_groups:
        rois = []
        roi_attrs = []
        roi_names = []
        roi_numbers = []
        roi_scan_origin = []
        for file in scan_groups[group]:
            data = h5py.File(file, "r+")
            if args.copyfrom is not None:
                if args.copyfrom not in data["raw_data"].attrs["name"]:
                    continue
            if "rois" in data:
                for a in data["rois"]:
                    deletion = False  # track whether any rois deleted so we can renumber at the end
                    for b in data["rois"][a]:
                        if data["rois"][a][b].attrs.get("copy", default=False):
                            if args.deleteold or args.justdeletecopies:
                                del data["rois"][a][b]
                                deletion = True
                            continue
                        rois.append(data["rois"][a][b][:])
                        roi_attrs.append(dict(data["rois"][a][b].attrs))
                        roi_names.append(a)
                        roi_numbers.append(b)
                        roi_scan_origin.append(data["raw_data"].attrs["name"])
                    if deletion:
                        original = sorted(list(data["rois"][a].keys()), key=int)
                        change = [str(x) for x in range(len(original))]
                        for old, new in zip(original, change):
                            data["rois"][a].move(old, new)
            data.close()
        for file in scan_groups[group]:
            data = h5py.File(file, "r+")
            if args.copyto is not None:
                if args.copyto not in data["raw_data"].attrs["name"]:
                    continue
            for roi, att, nam, num, ori in zip(rois, roi_attrs, roi_names, roi_numbers, roi_scan_origin):
                if ori != data["raw_data"].attrs["name"]:
                    zs = np.unique(data["Z-POS"][:])
                    make_copy = False
                    if att["z"] in zs:
                        # print("Able to copy from", ori, data["raw_data"].attrs["name"])
                        make_copy = True
                    elif np.any(np.isclose(zs, att["z"], atol=1, rtol=0)):
                        print("Not exact, but close for: ", ori, "to", data["raw_data"].attrs["name"])
                        copy = input("Not exact match, copy anyway? Y/[N]") if not args.copyclose else True
                        if copy in ["Y", "y", True]:
                            make_copy = True
                            old_z = att["z"]
                            att["z"] = zs[np.argmin(np.abs(zs - att["z"]))]
                            print("Copying, changing z from {:.2f} to {:.2f}".format(old_z, att["z"]))
                    else:
                        # print("No close positions for: ", ori, "to", data["raw_data"].attrs["name"], "so skipping.")
                        # print(zs[np.argmin(np.abs(zs - att["z"]))], att["z"])
                        # print(np.isclose(zs, att["z"], atol=0.1, rtol=0))
                        make_copy = False
                    if make_copy and not args.justdeletecopies:
                        # copy the roi to scan
                        roi_grp = data.require_group("rois")
                        roi_grp = roi_grp.require_group(nam)
                        roi_num = str(len(roi_grp.keys()))
                        roi_data = roi_grp.create_dataset(roi_num, data=roi)
                        roi_data.attrs["copy"] = True
                        for a in att:
                            roi_data.attrs[a] = att[a]
            data.close()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import logging
from abc import ABC, abstractmethod
from typing import Sequence, Tuple, Optional, List

import numpy as np

from ..core.image_structures.pa_time_data import PATimeSeries
from ..core.image_structures.reconstruction_image import Reconstruction
from ..io.attribute_tags import ReconAttributeTags
from ..io.msot_data import PAData, HDF5Tags
from ..processing.processing_algorithm import TimeSeriesProcessingAlgorithm, ProcessingResult


class ReconstructionAlgorithm(TimeSeriesProcessingAlgorithm, ABC):
    def pre_prepare_data(self, x):
        return x

    def __init__(self,
                 n_pixels: Sequence[int],
                 field_of_view: Sequence[float],
                 **kwargs):
        super().__init__()
        self.n_pixels = n_pixels
        self.field_of_view = field_of_view
        self.custom_params = kwargs

    @abstractmethod
    def reconstruct(self, raw_data: np.ndarray,
                    fs: float,
                    geometry: np.ndarray,
                    n_pixels: Sequence[int],
                    field_of_view: Sequence[float],
                    speed_of_sound,
                    **kwargs) -> np.ndarray:
        pass

    @staticmethod
    def get_algorithm_name() -> str:
        pass

    def run(self, time_series: PATimeSeries,
            pa_data: PAData,
            speed_of_sound=None,
            geometry=None,
            **kwargs) -> Tuple[Reconstruction, dict, Optional[List[ProcessingResult]]]:
        from .. import PAT_MAXIMUM_BATCH_SIZE
        speed_of_sound = pa_data.get_speed_of_sound() if speed_of_sound is None else speed_of_sound

        if geometry is None:
            geometry = pa_data.get_scan_geometry()

        logging.debug(f"{time_series.attributes}, {geometry.shape}, {time_series.raw_data.shape}, {speed_of_sound}")

        irf = pa_data.get_impulse_response() if pa_data is not None else None
        wavelengths = time_series.da.coords.get("wavelengths") if pa_data is None else pa_data.get_wavelengths()

        if type(self.field_of_view[0]) in [tuple, list, np.ndarray]:
            # Off centre field_of_view
            fov = [abs(x1 - x0) for (x1, x0) in self.field_of_view]
            geometry = np.array(geometry)
            for i in range(len(self.field_of_view)):
               geometry[:, i] -= np.mean(self.field_of_view[i])
        else:
            fov = self.field_of_view
        
        # Process in batches to avoid GPU running out of memory.
        if time_series.shape[0] * time_series.shape[1] > PAT_MAXIMUM_BATCH_SIZE != -1:
            new_recons = []
            ts_raw = time_series.raw_data
            shape = ts_raw.shape
            ts_raw = ts_raw.reshape((-1,) + shape[-2:])
            for i in range(0, ts_raw.shape[0], PAT_MAXIMUM_BATCH_SIZE):
                raw = self.reconstruct(ts_raw[i:i + PAT_MAXIMUM_BATCH_SIZE], time_series.attributes["fs"],
                                       geometry, self.n_pixels, fov, speed_of_sound,
                                       irf=irf, **kwargs, **self.custom_params)
                new_recons.append(np.asarray(raw))
            raw_data = np.concatenate(new_recons, axis=0).reshape(shape[:2] + new_recons[0].shape[1:])
        else:
            raw_data = self.reconstruct(time_series.raw_data,
                                        time_series.attributes["fs"],
                                        geometry,
                                        self.n_pixels,
                                        fov,
                                        speed_of_sound,
                                        irf=irf,
                                        **kwargs, **self.custom_params)

        output_data = Reconstruction(raw_data, wavelengths,
                                     hdf5_sub_name=self.get_algorithm_name(),
                                     field_of_view=self.field_of_view)
        output_data.attributes[HDF5Tags.SPEED_OF_SOUND] = speed_of_sound
        output_data.attributes[ReconAttributeTags.RECONSTRUCTION_ALGORITHM] = self.get_algorithm_name()
        output_data.attributes[ReconAttributeTags.X_NUMBER_OF_PIXELS] = self.n_pixels[0]
        output_data.attributes[ReconAttributeTags.Y_NUMBER_OF_PIXELS] = self.n_pixels[1]
        output_data.attributes[ReconAttributeTags.Z_NUMBER_OF_PIXELS] = self.n_pixels[2]
        output_data.attributes[ReconAttributeTags.X_FIELD_OF_VIEW] = self.field_of_view[0]
        output_data.attributes[ReconAttributeTags.Y_FIELD_OF_VIEW] = self.field_of_view[1]
        output_data.attributes[ReconAttributeTags.Z_FIELD_OF_VIEW] = self.field_of_view[2]
        output_data.attributes[ReconAttributeTags.ADDITIONAL_PARAMETERS] = kwargs
        output_data.attributes[HDF5Tags.WAVELENGTH] = wavelengths

        for a in time_series.attributes:
            output_data.attributes[a] = time_series.attributes[a]

        return output_data, {}, None
